# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
        Classes for the simulation of neutral and charged particles in a plasma discharge.

        This module contains a class for managing the neutral particles (mainly atoms and molecules) 
        which are the targets of collisions by moving particles, usually electrons and ions.
        These target particles are considered fixed in space, since their speed is much lower respect to moving particles.
        Cross sections needed to treat the collision processes are imported from ASCII files.

        How collisions are treated
        --------------------------

        To decide if an electron (or ion) scatters whith a neutral (by any type of scattering process: elastic, 
        ionization, dissociation of any type) during a time interval dt, we consider that the 
        probability for (at least) a scattering event to happen during a time interval dt is
        p = 1 - exp( f(v) * dt ) = 1 - exp( r(v) * v * dt ) = 1 - exp( s(v) * n * v * dt )
        where: 
                v           -> electron speed
                n           -> total neutral density (sum of densities of all neutral types)
                f(v)=r(v)*v -> scattering frequency (number of scattering events experienced by a single electron in a unit time)
                r(v)=s(v)*n -> total scattering rate (number of scattering events for a unit length of 
                               path travelled by the electron)
                s(v)        -> total cross section (sum of all cross sections for all scattering types) 
                               at speed v

        Once decided that a scattering has happened, to decide of which type it was, we have to use
        relative probabilities and probability limits.
        The relative probability p_i that the scattering was of type i is
                p_i = r_i(v)/r(v)
        where:
                i      -> identifies the type of scattering process
                          0                <= i < self.types-1      -> elastic scattering (with neutral type i)
                          self.types       <= i < 2*self.types-1    -> ionization 
                                                                       (of neutral type i-self.types)
                          2*self.types     <= i < 2*self.types   
                                                  + n_excitations   -> excitation
                                                                       (neutral type and dissociation
                                                                       process depending on the value of i)
                          2*self.types-1 +     
                          + n_excitations  <= i                     -> dissociation 
                                                                       (neutral type and dissociation
                                                                       process depending on the value of i)

                r_i(v) -> scattering rate for process i at speed v
                r(v)   -> total scattering rate for all processes at speed v
                of course, the sum of all p_i gives 1.
                To decide the type of scattering, we extract a random numbed rand between 0 and 1, than compare the
                        number with the limits PL_i defined as follows
                        PL_0 = p_0
                        PL_i = PL_(i-1) + p_i           for i>0
                The type of scattering is given by the value of i for which it is true that
                        PL_(i-1) < rand < PL_i          whith i>0
                        or
                        0        < rand < PL_0          if i=0

        Recombination
        -------------

        The same considerations hold for recombination of electrons with ions, by the three main recombination channels:
        dissociative recombination, three body recombination and radiative recombination.
        In this case, we must substuitute the neutral density n with the ion density n_ion. However, while the former
        can be considered approximately constant if the ionization degree is very low, the latter is changing in time
        as the ionization process proceeds.

        Documentation is also available in the docstrings.
"""



# +--------------------------------+
# | Import required Python modules |
# +--------------------------------+

import math
import numpy

from pysica.parameters import *
from pysica.constants import *
from pysica.plasmapro.ccpla_defaults import *     
from pysica.functions.physics import number_density, mean_speed_maxwell
from pysica.managers.data_manager import DataGrid
from pysica.managers.gnuplot_manager import gnuplot_installed, new_plot, plot2d, plot_curves, plot_close, plot_close_all
from pysica.io.io_files import write_to_disk
from pysica.plasmapro.discharge.particles_data_manager import read_file_ntypes, read_file_properties


XSEC_PLOT_PERSIST = True

# +------------------+
# | Target particles |
# +------------------+

class TargetParticles:
        """A class defining a collection of particles, to be considered as the targets of collisions by electrons and ions."""

        def __init__(self, N_sigma_electrons, N_sigma_ions, temperature, total_pressure, min_scattered=DEFAULT_MIN_SCATTERED,
                     isactive_recomb=True, neutral_types=None, filename=None):
                """ Initialises the properties of target particles.

                        Initialises the properties of target particles.
                        If the number of target particle types is not specified, it will be read from a ASCII file containing
                        particles properties.

                        Parameters
                        ----------

                        N_sigma_electrons: number of energies for which cross section for electrons scattering values are tabulated
                        N_sigma_ions:      number of energies for which cross section for ions scatterig values are tabulated
                        temperature:       temperature of gases in K
                        total_pressure:    total pressure of gases in Pa
                        isactive_recomb:   set if electron/ion recombination processes must be activated or not
                        neutral_types:     number of neutral types, if this value is equal to None
                                              than the number of neutrals will be read from 
                                              the ASCII file with name filename
                        filename:          name of the ASCII file containing neutrals properties
                                              if neutral_types=None this parameter cannot be left blank

                        Initialized data attributes
                        ---------------------------

                        self.read_error:                used for error checking
                                                        (status, message)
                                                                status: 0 = no error
                                                                        1 = file not found
                                                                        2 = no gases were given in the file
                                                                message: an error message or 'OK'
                        self.temperature:               temperature of the gas mixture / K
                        self.total_pressure:            total pressure of the gas mixture / Pa
                        self.isactive_recomb:           set if electron/ion recombination processes must be activated or not
                        self.types:                     how many types of target molcules are involved
                        self.n_sigma:                   number of velocity values for which electron impact cross sections 
                                                               are tabulated
                        self.n_sigma_ions:              number of velocity values for which ion impact cross sections are tabulated
                """

                self.read_error = (0, 'OK')

                # If the number of gases was not given, try to read it from file
                if (neutral_types is None):
                        if (filename is None):
                                self.read_error = (3, "Filename was not given")
                                return
                        else:
                                self.read_error = read_file_ntypes(filename, self)
                                if (self.read_error[0] != 0): return
                else:
                        self.types              = neutral_types

                # Characteristics of the plasma discharge
                self.temperature                = temperature
                self.total_pressure             = total_pressure
                self.isactive_recomb            = isactive_recomb

                # Number of velocity values used to discretize cross sections for electron and ion impact with neutrals
                self.n_sigma                    = N_sigma_electrons
                self.n_sigma_ions               = N_sigma_ions
                self.min_scattered              = min_scattered


        # +-------------------------------------------------------+
        # | Get energies for which cross section values are known |
        # +-------------------------------------------------------+

        def sigma_energies(self):
                """Returns an array with the energies (in eV) for which cross section values for electrons are tabulated.

                        Returns
                        -------

                        array of the energies (expressed in eV) for which the cross section values are tabulated

                """

                return - 0.5 * ELECTRON_MASS / ELECTRON_CHARGE * self.sigma_velocities**2


        def sigma_energies_ions(self, ion_type):
                """Returns an array with the energies (in eV) for which cross section values for ions are tabulated.

                        Parameters
                        ----------

                        ion_type: the type of ion for the energies are requested


                        Returns
                        -------

                        array of the energies (expressed in eV) for which the cross section values are tabulated

                """

                return - 0.5 * self.mass[ion_type] * ATOMIC_UNIT_MASS / ELECTRON_CHARGE * self.sigma_velocities_ions[ion_type]**2


        # +--------------------------------+
        # | Get the properties of neutrals |
        # +--------------------------------+

        def read_properties(self, filename, sep, e_min_sigma, e_max_sigma, e_min_sigma_ions, e_max_sigma_ions, debug=False):
                """Reads the neutrals characteristics from an ascii file and calculates some properties.

                        Parameters
                        ----------

                        filename:         name of file containing neutral gases characteristics
                        sep:              character used in the ASCII file to separate values (usually the tab char)
                        e_min_sigma:      minimum energy for which cross sections for electron scattering are tabulated
                        e_max_sigma:      maximum energy for which cross sections for electron scattering are tabulated
                        e_min_sigma_ions: minimum energy for which cross sections for ion scattering are tabulated
                        e_max_sigma_ions: maximum energy for which cross sections for ion scattering are tabulated

                        Initialized data attributes
                        ---------------------------
                        self.names:                      names of the gas molecules (max 8 char)
                        self.molecule_type:              string describing the type of molecule (1 char)
                                                                'a'= atom, 'm'= nonpolar molecule, 'p'= polar molecule
                                                                actually polar and non-polar molecules are treated in the 
                                                                same way, but this may change in the future
                        self.mass:                       masses of gas molecules / u
                        self.mean_v:                     mean speeds of gas molecules (all at the same temperature)
                        self.gas_flow:                   array of gas fluxes / arbitrary units
                        self.partial_pressure:           array of gases partial pressures / Pa
                        self.number_density:             number density of neutral species
                        self.secondary_emission:         secondary emission coeff. i.e. average number of e- emitted 
                                                                when an ion is collected at one electrode
                        self.dissociation_rate           dissociation rate for each molecule type and its dissociation types
                        self.dissociation_rate_constant  dissociation rate constant for each molecule type and its dissociation types

                        self.energy_loss:                fraction of energy transferred from an electron to a neutral particle
                                                                during elastic scattering
                        self.energy_loss_ions:           fraction of energy transferred from an ion to a neutral particle
                                                                during elastic scattering
                        self.v_ratio:                    electron speed reduction during elastic scattering
                        self.v_ratio_ions:               ion speed reduction during elastic scattering
                        self.ionization_energy:          first ionization energy / eV
                        self.excitation_tpyes            number of excitation processes for each neutral type
                        self.excitation_energy:          excitation energies for processes to be considered / eV
                        self.dissociation_types:         number of dissociation processes for each neutral type
                        self.dissociation_energy:        dissociation energies for processes to be considered / eV

                        self.sigma_velocities:           array of velocity values for which electron impact cross sections 
                                                                are tabulated
                        self.sigma_velocities_ions:      matrix of velocity values for which cross sections are tabulated 
                                                                for each ion type

                        self.sigma_elastic:              cross section table for electron impact elastic scattering 
                        self.sigma_ionization:           cross section table for electron impact ionization
                        self.sigma_excitation            cross section table for electron impact excitation
                        self.sigma_dissociation:         cross section table for electron impact dissociation   

                        self.sigma_recombination_diss:   cross section table for electron impact dissociative recombination 
                                                                of ionized molecules
#                       self.sigma_recombination_3body:  cross section table for electron impact three body recombination 
#                                                                of ionized atoms/molecules
#                       self.sigma_recombination_rad:    cross section table for electron impact radiative recombination 
#                                                                of ionized atoms/molecules

                        self.sigma_elastic_ions:         cross section table for ion impact elastic scattering 
                        self.sigma_charge_exchange_ions: cross section table for ion impact elastic scattering 

                        self.collisions_null:            number of electron null collisions
                        self.collisions_elastic:         number of electron elastic collisions
                        self.collisions_ionization:      number of electron ionization collisions
                        self.collisions_excitation:      number of electron excitation collisions
                        self.collisions_dissociation:    number of electron dissociation collisions
                        self.collisions_recombination:   number of electron/ion recombination collisions
                        self.collisions_total_electron:  total number of electron collisions, of any type
                        self.dissociation_rate:          dissociation rate of neutral particles, 
                                                               for each type and dissociation process
                        self.dissociation_rate_constant: dissociation rate constant of neutral particles, 
                                                               for each type and dissociation process

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = could not open file
                                         2 = missing values
                                         3 = numerical value expected
                                         4 = unknown molecule type
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                # Characteristics of gases
                self.names                      = numpy.zeros(self.types, 'U8')
                self.molecule_type              = numpy.zeros(self.types, 'U1')
                self.mass                       = numpy.zeros(self.types, 'd')
                self.mean_v                     = numpy.zeros(self.types, 'd')
                self.gas_flow                   = numpy.zeros(self.types, 'd')
                self.partial_pressure           = numpy.zeros(self.types, 'd')
                self.number_density             = numpy.zeros(self.types, 'd')
                self.secondary_emission         = numpy.zeros(self.types, 'd')

                # Characteristics for interaction with electrons
                self.energy_loss                = numpy.zeros(self.types, 'd')
#                self.v_ratio                    = numpy.zeros(self.types, 'd')
                self.ionization_energy          = numpy.zeros(self.types, 'd')
                self.excitation_types           = numpy.zeros(self.types, 'b')
                self.excitation_energy          = numpy.zeros( (self.types, MAX_EXCITATION_TYPES), 'd', order='F' )
                self.dissociation_types         = numpy.zeros(self.types, 'b')
                self.dissociation_energy        = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES), 'd', order='F' )

                # Characteristics for interaction with ions
                self.energy_loss_ions           = numpy.zeros( (self.types, self.types), 'd', order='F' )
#                self.v_ratio_ions               = numpy.zeros( (self.types, self.types), 'd', order='F' )

                # Arrays of velocity values for which cross sections are tabulated
                self.sigma_velocities           = numpy.zeros(self.n_sigma, 'd')        
                self.sigma_velocities_ions      = numpy.zeros( (self.types, self.n_sigma_ions), 'd', order='F') 

                # Arrays of cross sections for electron impact processes
                self.sigma_elastic              = numpy.zeros( (self.types, self.n_sigma), 'd', order = 'F')
                self.sigma_ionization           = numpy.zeros( (self.types, self.n_sigma), 'd', order = 'F')
                self.sigma_excitation           = numpy.zeros( (self.types, MAX_EXCITATION_TYPES, self.n_sigma), 
                                                             'd', order = 'F')
                self.sigma_dissociation         = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES, self.n_sigma), 
                                                             'd', order = 'F')

                self.sigma_recombination_diss   = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES, self.n_sigma), 
                                                             'd', order = 'F')
                self.sigma_recombination_3body  = numpy.zeros( (self.types, self.n_sigma), 'd', order = 'F')
                self.sigma_recombination_rad    = numpy.zeros( (self.types, self.n_sigma), 'd', order = 'F')

                # Arrays of cross sections for ion impact processes
                self.sigma_elastic_ions         = numpy.zeros( (self.types, self.types, self.n_sigma_ions), 'd', order='F')
                self.sigma_charge_exchange_ions = numpy.zeros( (self.types, self.types, self.n_sigma_ions), 'd', order='F')

                # Quantities that will be calculated during the simulation
                # collision counters for electron/neutral collisions and electron/ion recombinations
                # and dissociation rate / rate constant counters
                self.collisions_null            = numpy.zeros(1, 'd') # Must be an array since it will be passed to f2py
                self.collisions_elastic         = numpy.zeros(self.types, 'd')
                self.collisions_ionization      = numpy.zeros(self.types, 'd')
                self.collisions_excitation      = numpy.zeros( (self.types, MAX_EXCITATION_TYPES),   'd', order='F' )
                self.collisions_dissociation    = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES), 'd', order='F' )
                self.collisions_recombination   = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES), 'd', order='F' )
                self.collisions_total_electron  = 0
                self.dissociation_rate          = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES), 'd', order='F' )
                self.dissociation_rate_constant = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES), 'd', order='F' )               

               
                # Read neutral properties from ascii file
                (status, message) = read_file_properties(filename, sep, self, debug)
                if (status != 0): return (status, message)

                # Calculate neutral properties
                self.partial_pressure   = self.gas_flow * self.total_pressure / self.gas_flow.sum()
                self.number_density     = number_density(self.partial_pressure, self.temperature)
                
                # Average electron energy lost during an elastic scattering
                self.energy_loss        = 2 * ELECTRON_MASS / (self.mass * ATOMIC_UNIT_MASS)
                # Average ratio of electron speed measured after and before an elastic scattering
#                self.v_ratio            = numpy.sqrt(numpy.abs(1-self.energy_loss))                     
                self.mean_v             = mean_speed_maxwell(self.temperature, self.mass)

                for i in range(self.types):
                        for j in range(self.types):
                                # Average ion energy lost during an elastic scattering
                                self.energy_loss_ions[i][j] = 2 * self.mass[i] * self.mass[j] / (self.mass[i] + self.mass[j])**2
                                # Average ratio of ion speed measured after and before an elastic scattering
#                                self.v_ratio_ions[i][j]     = numpy.sqrt(1-self.energy_loss_ions[i][j])

                # Calculate speed values at which cross sections for electrons will be interpolated
                v_min_sigma = math.sqrt( 2 * e_min_sigma * (-ELECTRON_CHARGE / ELECTRON_MASS) )
                v_max_sigma = math.sqrt( 2 * e_max_sigma * (-ELECTRON_CHARGE / ELECTRON_MASS) )
                self.sigma_velocities = numpy.linspace(v_min_sigma, v_max_sigma, self.n_sigma )

                # Calculate speed values at which cross sections for each ion type will be interpolated
                # kinetic energy values are stored in eV, so they are multiplied by electron charge
                # to transform them in joule (ELECTRON_CHARGE is a negative number)
                for i in range(self.types):
                        v_min_sigma = math.sqrt( 2 * e_min_sigma_ions * (-ELECTRON_CHARGE / (self.mass[i] * ATOMIC_UNIT_MASS)) )
                        v_max_sigma = math.sqrt( 2 * e_max_sigma_ions * (-ELECTRON_CHARGE / (self.mass[i] * ATOMIC_UNIT_MASS)) )
                        self.sigma_velocities_ions[i] = numpy.linspace(v_min_sigma, v_max_sigma, self.n_sigma_ions)

                return (status, message)


        # +--------------------------------------------------------------------+
        # | General method used to read cross section values from a ascii file |
        # +--------------------------------------------------------------------+

        def _read_xsec(self, filename, sep, qm_ratio, target_index, ion_index=None, plot=False, title=''):
                """Reads a table of cross section values as a function of energy from a ascii file.

                        Parameters
                        ----------

                        filename:       name of file containing cross section values at energies in the form
                                                #Comment
                                                <energy> <separator> <value>    #Comment
                                                <energy> <separator> <value>    #Comment
                                        where energies are in eV and cross section values are in m**2
                                        
                        sep:            character used to separate values (usually the tab char)

                        qm_ratio:       the charge to mass ratio of incident particle

                        ion_index:      integer value indentifying the type of ion (i.e. the type of neutral that was ionized) 
                                        that scatters with the target, 
                                        if it is None, than the particle is an electron

                        target_index:   integer value indentifying the type of target

                        plot:           if True, plot the cross section values as a function of energy,
                                        before and after interpolation

                        title:          a string containing the title to be printed on the plot


                        Initialized data attributes
                        ---------------------------

                        self.sigma_return       numpy array containing the cross section values that were read from the file


                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = negative energy value found
                                         2 = negative cross-section value found
                                         3 = ion index out of bounds
                                message: a string containing an error message or 'Ok'
                """


                status, message  = 0, 'OK'              # This variable is used for error checking      

                if (ion_index is None):
                        self.sigma_return= numpy.zeros(self.n_sigma, 'd')
                else:
                        if ( (ion_index < 0) or (ion_index > self.types-1)):
                                status = 3
                                message = 'ion type ' + str(ion_index) + ' is out of bounds'
                                return(status, message)                                
                        self.sigma_return= numpy.zeros(self.n_sigma_ions, 'd')

                # Read the tabulated values of elastic cross-section into a numpy array
                sigma = DataGrid()
                (status, message) = sigma.read_file(filename, n_columns=2, sep=sep, transpose=True)
                if (status != 0):
                        return (status, message)
                else: 
                        sigma.n_data = len(sigma.data_array[0])

                # Get the absolute value of mass to charge ratio
                if (qm_ratio < 0): qm_ratio = - qm_ratio

                # Check that all cross-section values are positive
                for j in range(sigma.n_data):
                        if sigma.data_array[0][j] < 0:  
                                status  = 1
                                message = 'negative energy value found in row # '+str(j+1)
                                return (status, message)
                        if sigma.data_array[1][j] < 0:
                                status  = 2
                                message = 'negative cross-section value found in row # '+str(j+1)
                                return (status, message)

                # Change from energies [eV] to velocities [m/s]
                sigma.data_array[0] = numpy.sqrt( 2 * qm_ratio * sigma.data_array[0] )

                # Calculate a interpolated cross-section table
                if (ion_index is None): 
                        self.sigma_return = numpy.interp(self.sigma_velocities,
                                                         sigma.data_array[0],
                                                         sigma.data_array[1])
                else:
                        self.sigma_return = numpy.interp(self.sigma_velocities_ions[ion_index],
                                                         sigma.data_array[0],
                                                         sigma.data_array[1])

                # Plot cross-section
                if (plot and gnuplot_installed): 
                        xsec_graph = new_plot(title=title,
                                              xlabel='Energy / eV',
                                              ylabel='Cross section / m**2',
                                              format_x='%g',
                                              format_y='%g',
                                              logx=True,
                                              grid=True,
                                              persistence=XSEC_PLOT_PERSIST,
                                              redirect_output=True)
                       
                        data_list = []
                        
                        # These are the values given in the data file
                        data_list.append( [0.5 / qm_ratio * sigma.data_array[0]**2,
                                                sigma.data_array[1],
                                                'Tabulated values',
                                                None ] )
                        
                        # These are the values calculated by interpolation of the ones in the data file
                        if (ion_index is None):
                                data_list.append( [ self.sigma_energies(),
                                                         self.sigma_return,
                                                         'Interpolated values',
                                                         None ] )
                        else:
                                data_list.append( [ self.sigma_energies_ions(ion_index),
                                                         self.sigma_return,
                                                         'Interpolated values',
                                                         None ] )
                       
                        plot_curves(xsec_graph, data_list, volatile=True)
                       
                        # Close the plots, to quit the associated gnuplot processes
                        # plots remain visible thanks to the "persist" option
                        plot_close(xsec_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)

                        del(xsec_graph)

                return (status, message)


        # +------------------------------------------------------------------+
        # | Methods to read cross sections for electrons/neutrals scattering |
        # +------------------------------------------------------------------+

        def read_xsec_electrons_elastic(self, filename, sep, neutral_index, plot=False):
                """Reads the table of elastic cross section values as a function of energy from a ascii file.

                        Parameters
                        ----------

                        filename:       name of file containing cross section values at energies in the form

                                                #Comment
                                                <energy> <separator> <value>    #Comment
                                                <energy> <separator> <value>    #Comment

                                        where energies are in eV and cross section values are in m**2
                                        
                        sep:            character used to separate values (usually the tab char)
                        neutral_index:  integer value indentifying the type of neutral for which cross section
                                        shuld be read

                        plot:           if True, plot the cross section values as a function of energy,
                                        before and after interpolation


                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = negative energy value found
                                         2 = negative cross-section value found
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                title    = 'Cross section for elastic scattering of e- on '+self.names[neutral_index]
                qm_ratio = - ELECTRON_CHARGE / ELECTRON_MASS

                (status, message) = self._read_xsec(filename, sep, qm_ratio, neutral_index, None, plot, title)

                if (status == 0): self.sigma_elastic[neutral_index] = self.sigma_return

                return (status, message)


        def read_xsec_electrons_ionization(self, filename, sep, neutral_index, check=False, plot=False):
                """Reads the table of ionization cross section values as a function of energy from a ascii file.

                        Parameters
                        ----------

                        filename:       name of file containing cross section values at energies in the form

                                                        #Comment
                                                        <energy> <separator> <value>    #Comment
                                                        <energy> <separator> <value>    #Comment

                                        where energies are in eV and cross section values are in m**2
                                                        

                        sep:            character used to separate values (usually the tab char)

                        neutral_index:  integer value indentifying the type of neutral for which cross section
                                        shuld be read

                        check:          if True, check that all cross section values for energies lower
                                        than ionization threshold are zero

                        plot:           if True, plot the cross section values as a function of energy,
                                        before and after interpolation


                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = negative energy value found
                                         2 = negative cross-section value found
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                title    = 'Cross section for e- impact ionization of '+self.names[neutral_index]
                qm_ratio = - ELECTRON_CHARGE / ELECTRON_MASS

                (status, message) = self._read_xsec(filename, sep, qm_ratio, neutral_index, None, plot, title)

                if (status == 0): self.sigma_ionization[neutral_index] = self.sigma_return

                # Check that cross section is zero for energies lower than ionization threshold
                if check:
                        # Calculate speed corresponding to ionization treshold
                        v_min = math.sqrt( 2.0 * self.ionization_energy[neutral_index] * qm_ratio )
                        i=0
                        while(self.sigma_velocities[i]<=v_min): 
                                self.sigma_ionization[neutral_index, i]=0
                                i=i+1

                return (status, message)


        def read_xsec_electrons_excitation(self, filename, sep, neutral_index, exc_type, check=False, plot=False):
                """Reads the table of excitation cross section values as a function of energy from a ascii file.

                        Parameters
                        ----------

                        filename:       name of the file containing cross section values at energies in the form

                                                #Comment
                                                <energy> <separator> <value>    #Comment
                                                <energy> <separator> <value>    #Comment

                                        where energies are in eV and cross section values are in m**2
                                                

                        sep:            character used to separate values (usually the tab char)

                        neutral_index:  integer value indentifying the type of neutral for which cross section
                                                shuld be read

                        exc_type:       integer value identifying the type of excitation process

                        check:          if True, check that all cross section values for energies lower
                                        than dissociation threshold are zero

                        plot:           if True, plot the cross section values as a function of energy,
                                                before and after interpolation

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = negative energy value found
                                         2 = negative cross-section value found
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                title    = 'Cross section for e- impact excitation of '+ self.names[neutral_index]+' process #'+str(exc_type)
                qm_ratio = - ELECTRON_CHARGE / ELECTRON_MASS

                (status, message) = self._read_xsec(filename, sep, qm_ratio, neutral_index, None, plot, title)

                if (status == 0): self.sigma_excitation[neutral_index, exc_type] = self.sigma_return

                # Check that cross section is zero for energies lower than excitation threshold
                if check:
                        # Calculate speed corresponding to excitation treshold
                        v_min = math.sqrt( 2.0 * self.excitation_energy[neutral_index, exc_type] * qm_ratio )
                        i=0
                        while(self.sigma_velocities[i]<=v_min): 
                                self.sigma_excitation[neutral_index, exc_type, i]=0
                                i=i+1                   

                return (status, message)
        
        

        def read_xsec_electrons_dissociation(self, filename, sep, neutral_index, diss_type, check=False, plot=False):
                """Reads the table of dissociation cross section values as a function of energy from a ascii file.

                        Parameters
                        ----------

                        filename:       name of the file containing cross section values at energies in the form

                                                #Comment
                                                <energy> <separator> <value>    #Comment
                                                <energy> <separator> <value>    #Comment

                                        where energies are in eV and cross section values are in m**2
                                                

                        sep:            character used to separate values (usually the tab char)

                        neutral_index:  integer value indentifying the type of neutral for which cross section
                                                        shuld be read

                        diss_type:      integer value identifying the type of dissociation process

                        check:          if True, check that all cross section values for energies lower
                                        than dissociation threshold are zero

                        plot:           if True, plot the cross section values as a function of energy,
                                                before and after interpolation

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = negative energy value found
                                         2 = negative cross-section value found
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                title    = 'Cross section for e- impact dissociation of '+ self.names[neutral_index]+' process #'+str(diss_type)
                qm_ratio = - ELECTRON_CHARGE / ELECTRON_MASS

                (status, message) = self._read_xsec(filename, sep, qm_ratio, neutral_index, None, plot, title)

                if (status == 0): self.sigma_dissociation[neutral_index, diss_type] =self.sigma_return

                # Check that cross section is zero for energies lower than dissociation threshold
                if check:
                        # Calculate speed corresponding to dissociation treshold
                        v_min = math.sqrt( 2.0 * self.dissociation_energy[neutral_index, diss_type] * qm_ratio )                      
                        i=0
                        while(self.sigma_velocities[i]<=v_min): 
                                self.sigma_dissociation[neutral_index, diss_type, i]=0
                                i=i+1                   

                return (status, message)


        def calculate_total_xsec_electrons(self, debug=False):
                """Calculates main statistical parameters from cross sections.

                        Calculate, for the different processes (elastic scattering, ionization, dissociation types): 
                                - total cross sections
                                - collision frequencies
                                - relative scattering probabilities
                                - probability boundaries
                        (see the module docstring for more information on these quantities and their usage)

                        Parameters
                        ----------

                        Initialized data attributes
                        ---------------------------

                        self.sigma_total_elastic:       sum of elastic scattering cross sections for all neutral types
                        self.sigma_total_ionization:    sum of ionization cross sections for all neutral types
                        self.sigma_total_excitation:    sum of excitation cross sections for all neutral types
                                                                and all excitation processes for each neutral type
                        self.sigma_total_dissociation:  sum of dissociation cross sections for all neutral types
                                                                and all dissociation processes for each neutral type
                        self.sigma_total_global:        sum of all cross section for all processes

                        self.frequency_elastic:         collision frequencies for elastic scattering
                        self.frequency_ionization:      collision frequencies for ionization scattering
                        self.frequency_excitation:      collision frequencies for excitation scattering
                        self.frequency_dissociation:    collision frequencies for dissociation scattering
                        self.frequency_total_global:    sum of all collision frequencies, for all prossible processes
                        self.frequency_null_collision:  null collision frequency, defined as the difference between
                                                                the maximum value (over all velocities) of the 
                                                                total collision frequency
                                                                and the collision frequency

                        self.p_relative_elastic:        relative probability of elastic scattering
                        self.p_relative_ionization:     relative probability of ionization
                        self.p_relative_excitation:     relative probabilities excitation processes
                        self.p_relative_dissociation:   relative probabilities dissociation processes
                        self.p_relative_null_collision: relative probability of null collision                                  

                        self.nmax_limits:               maximum allowed number of probability limits
                                                        i.e. max number of processes
                        self.probability_limits:        limits for calculation of scattering type
                        self.probability_processes:     names of the processed associated to limits
                        self.probability_index_exc      index of the first probability limit for excitation
                        self.probability_index_diss     index of the first probability limit for dissociation

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = error
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                self.sigma_total_elastic        = numpy.zeros( self.n_sigma, 'd')
                self.sigma_total_ionization     = numpy.zeros( self.n_sigma, 'd')
                self.sigma_total_excitation     = numpy.zeros( self.n_sigma, 'd')
                self.sigma_total_dissociation   = numpy.zeros( self.n_sigma, 'd')
                self.sigma_total_global         = numpy.zeros( self.n_sigma, 'd')

                self.frequency_elastic          = numpy.zeros( (self.types, self.n_sigma), 'd', order = 'F')
                self.frequency_ionization       = numpy.zeros( (self.types, self.n_sigma), 'd', order = 'F')
                self.frequency_excitation       = numpy.zeros( (self.types, MAX_EXCITATION_TYPES, self.n_sigma), 
                                                               'd', order = 'F')                
                self.frequency_dissociation     = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES, self.n_sigma), 
                                                               'd', order = 'F')

                self.frequency_total_global     = numpy.zeros( self.n_sigma, 'd')
                self.frequency_null_collision   = numpy.zeros( self.n_sigma, 'd')

                self.p_relative_elastic         = numpy.zeros( (self.types, self.n_sigma), 'd', order='F')
                self.p_relative_ionization      = numpy.zeros( (self.types, self.n_sigma), 'd', order='F')
                self.p_relative_excitation      = numpy.zeros( (self.types, MAX_EXCITATION_TYPES, self.n_sigma), 
                                                                'd', order = 'F')                
                self.p_relative_dissociation    = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES, self.n_sigma), 
                                                                'd', order = 'F')
                self.p_relative_null_collision  = numpy.zeros( self.n_sigma, 'd')

                self.nmax_limits                = self.types * (2 + MAX_EXCITATION_TYPES + MAX_DISSOCIATION_TYPES)
                self.probability_limits         = numpy.zeros( (self.nmax_limits, self.n_sigma), 'd', order = 'F')
                self.probability_processes      = numpy.zeros( self.nmax_limits, 'U24')

                # Calculate total cross sections for all processes
                for i in range(self.types):
                        self.sigma_total_elastic    = self.sigma_total_elastic    + self.sigma_elastic[i]
                        self.sigma_total_ionization = self.sigma_total_ionization + self.sigma_ionization[i]
                        for j in range(self.excitation_types[i]):
                                self.sigma_total_excitation = ( self.sigma_total_excitation
                                                                + self.sigma_excitation[i,j] )
                        for j in range(self.dissociation_types[i]):
                                self.sigma_total_dissociation = ( self.sigma_total_dissociation
                                                                  + self.sigma_dissociation[i,j] )

                # Calculate total global cross-section
                self.sigma_total_global = (self.sigma_total_elastic      + self.sigma_total_ionization
                                           + self.sigma_total_excitation + self.sigma_total_dissociation  )

                # Calculate collision frequencies for all processes
                # f_i(v) = n_i * s_i(v) * v
                for i in range(self.types):
                        self.frequency_elastic[i]    = self.number_density[i] * self.sigma_elastic[i]    * self.sigma_velocities 
                        self.frequency_ionization[i] = self.number_density[i] * self.sigma_ionization[i] * self.sigma_velocities
                        for j in range(self.excitation_types[i]):
                                self.frequency_excitation[i,j] = ( self.number_density[i]
                                                                   * self.sigma_excitation[i,j]
                                                                   * self.sigma_velocities )
                        for j in range(self.dissociation_types[i]):
                                self.frequency_dissociation[i,j] = ( self.number_density[i]
                                                                     * self.sigma_dissociation[i,j]
                                                                     * self.sigma_velocities )

                # Calculate total collision frequency 
                # f_tot(v) = sum{f_i(v), i};  f_i(v) = n_i * s_i(v) * v
                for i in range(self.types):
                        self.frequency_total_global = ( self.frequency_total_global
                                                        + self.frequency_elastic[i]
                                                        + self.frequency_ionization[i] )
                        for j in range(self.excitation_types[i]):
                                self.frequency_total_global = ( self.frequency_total_global
                                                                + self.frequency_excitation[i,j] )                        
                        for j in range(self.dissociation_types[i]):
                                self.frequency_total_global = ( self.frequency_total_global
                                                                + self.frequency_dissociation[i,j] )

                # Calculate the maximum value of the total collision frequency over all velocities
                #   f_max = max{f_tot(v), v}
                self.frequency_total_global_max = self.frequency_total_global.max()

                # Calculate null collision frequency (required to apply the null collision method)
                #   f_null(v) = f_max - f_tot(v)
                self.frequency_null_collision = self.frequency_total_global_max - self.frequency_total_global

                # Calculate relative collision probabilities
                # prel_i(v) = f_i(v) / f_max
                self.p_relative_null_collision = self.frequency_null_collision / self.frequency_total_global_max
                for i in range(self.types):
                        self.p_relative_elastic[i]    = self.frequency_elastic[i]    / self.frequency_total_global_max
                        self.p_relative_ionization[i] = self.frequency_ionization[i] / self.frequency_total_global_max
                        for j in range(self.excitation_types[i]):
                                self.p_relative_excitation[i,j] = ( self.frequency_excitation[i,j]
                                                                    / self.frequency_total_global_max )
                        for j in range(self.dissociation_types[i]):
                                self.p_relative_dissociation[i,j] = ( self.frequency_dissociation[i,j]
                                                                      / self.frequency_total_global_max )
                                
                # Calculate probability boundaries                        
                self.probability_limits[0]    = self.p_relative_elastic[0]
                self.probability_processes[0] = 'el->'+ self.names[0] + ': elastic' 
                for i in range(1, self.types):
                        self.probability_limits[i] = ( self.probability_limits[i-1]
                                                       + self.p_relative_elastic[i] )
                        self.probability_processes[i] = 'el->'+ self.names[i] + ': elastic' 
                for i in range(self.types):
                        self.probability_limits[self.types+i] = ( self.probability_limits[self.types+i-1]
                                                                  + self.p_relative_ionization[i] )
                        self.probability_processes[self.types+i] = 'el->'+ self.names[i] + ': ionization'                       
                self.probability_index_exc = 2 * self.types
                i = self.probability_index_exc  # self.types*2
                
                for j in range(self.types):
                        for k in range(self.excitation_types[j]):
                                self.probability_limits[i] = ( self.probability_limits[i-1]
                                                               + self.p_relative_excitation[j,k] )
                                self.probability_processes[i] = 'el->'+ self.names[j] + ': excitation#' + str(k)                               
                                i = i + 1
                self.probability_index_diss = i
                for j in range(self.types):
                        for k in range(self.dissociation_types[j]):
                                self.probability_limits[i] = ( self.probability_limits[i-1]
                                                               + self.p_relative_dissociation[j,k] )
                                self.probability_processes[i] = 'el->'+ self.names[j] + ': dissociation#' + str(k)
                                i = i + 1                                
                self.n_limits = i
                if debug:
                        print('Probability boundaries for electron scattering')
                        for i in range(self.n_limits):
                                print(i, self.probability_processes[i])
        
                return (status, message)


        def plot_xsec_electrons(self, plot_single=True, plot_total=False,
                                plot_frequencies=False, plot_relative=False, plot_boundaries=False,
                                dot_points=False):
                """ Plots cross sections for electrons collisions and related main statistical parameters, 
                    as a function of electron energy.

                        Plots, for the different processes (elastic scattering, ionization, dissociation types): 
                                - total cross sections
                                - collision frequencies
                                - relative scattering probabilities
                                - probability boundaries

                        Parameters
                        ----------

                        plot_single:      if True, plot the cross section values of single processes as a function of energy
                        plot_total:       if True, plot the total cross section values as a function of energy
                        plot_frequencies: if True, plot the collision frequency values as a function of energy
                        plot_relative:    if True, plot the relative scattering probabilities
                        plot_boundaries:  if True, plot the probability boundaries
                        dot_points:       if True, plot using dots instead of markers

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = error
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking

                if (not gnuplot_installed): return

                if dot_points: plot_style = 'dots'
                else:          plot_style = 'points'

                # Plot cross-sections for single processes
                if plot_single:
                        data_list = []
                        # Prepare graph
                        self.cross_section_graph = new_plot(title='Cross sections for electron scattering',
                                                            xlabel='Energy / eV',
                                                            ylabel='Cross section / m**2',
                                                            logx=True,
                                                            logy=True,
                                                            format_x='%g',
                                                            format_y='%g',
                                                            grid=True,
                                                            style=plot_style,
                                                            persistence=XSEC_PLOT_PERSIST,
                                                            redirect_output=True)

                        # Plot cross sections for each process type and gas type
                        for i in range(self.types):
                                data_list.append( [ self.sigma_energies(),
                                                         self.sigma_elastic[i],
                                                         'elastic ' + self.names[i],
                                                         None ] )
                        for i in range(self.types):
                                data_list.append( [ self.sigma_energies(),
                                                         self.sigma_ionization[i],
                                                         'ionization ' + self.names[i],
                                                         None ] )
                        for i in range(self.types):
                                for j in range(self.excitation_types[i]):
                                        data_list.append( [ self.sigma_energies(),
                                                                 self.sigma_excitation[i,j],
                                                                 'excitation ' + self.names[i] + ' # ' + str(j),
                                                                 None ] )                        
                        for i in range(self.types):
                                for j in range(self.dissociation_types[i]):
                                        data_list.append( [ self.sigma_energies(),
                                                                 self.sigma_dissociation[i,j],
                                                                 'dissociation ' + self.names[i] + ' # ' + str(j),
                                                                 None ] )

                        plot_curves(self.cross_section_graph, data_list)

                        # Shut down gnuplot process, window remain visible due to persist option
                        plot_close(self.cross_section_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.cross_section_graph)

                
                # Plot total cross-sections
                if plot_total:
                        data_list = []
                        # Prepare graph
                        self.cross_section_graph = new_plot(title='Total cross sections for electron scattering',
                                                            xlabel='Energy / eV',
                                                            ylabel='Cross section / m**2',
                                                            logx=True,
                                                            logy=True,
                                                            format_x='%g',
                                                            format_y='%g',
                                                            grid=True,
                                                            style=plot_style,
                                                            persistence=XSEC_PLOT_PERSIST,
                                                            redirect_output=True)
                        # Plot total global cross section
                        data_list.append( [ self.sigma_energies(),
                                                 self.sigma_total_global, 
                                                 'total global',
                                                 None ] )
                        # Plot total cross sections for each process type
                        data_list.append( [ self.sigma_energies(),
                                                 self.sigma_total_elastic,
                                                 'total elastic',
                                                 None ] )
                        data_list.append( [ self.sigma_energies(),
                                                 self.sigma_total_ionization,
                                                 'total ionization',
                                                 None] )
                        data_list.append( [ self.sigma_energies(),
                                                 self.sigma_total_excitation,
                                                 'total excitation',
                                                 None ] )                        
                        data_list.append( [ self.sigma_energies(),
                                                 self.sigma_total_dissociation,
                                                 'total dissociation',
                                                 None ] )

                        plot_curves(self.cross_section_graph, data_list)
                        
                        # Shut down gnuplot process, window remain visible due to persist option
                        plot_close(self.cross_section_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.cross_section_graph)
                
                # Plot collision frequencies                
                if plot_frequencies:
                        data_list = []
                        self.coll_freq_graph = new_plot(title='Collision frequencies for electron scattering',
                                                        xlabel='Energy / eV',
                                                        ylabel='Frequency / Hz',
                                                        format_x='%g',
                                                        format_y='%g',
                                                        grid=True,
                                                        style=plot_style,
                                                        persistence=XSEC_PLOT_PERSIST,
                                                        redirect_output=True)                        
                        for i in range(self.types):
                                data_list.append( [ self.sigma_energies(),
                                                         self.frequency_elastic[i],
                                                         'elastic ' + self.names[i],
                                                         None ] )
                                data_list.append( [ self.sigma_energies(),
                                                         self.frequency_ionization[i],
                                                         'ionization ' + self.names[i],
                                                         None ] )
                                for j in range(self.excitation_types[i]):
                                        data_list.append( [ self.sigma_energies(),
                                                                 self.frequency_excitation[i,j],
                                                                 'excitation ' + self.names[i] + ' # ' + str(j),
                                                                 None ] )                                
                                for j in range(self.dissociation_types[i]):
                                        data_list.append( [ self.sigma_energies(),
                                                                 self.frequency_dissociation[i,j],
                                                                 'dissociation ' + self.names[i] + ' # ' + str(j),
                                                                 None ] )
                        data_list.append( [ self.sigma_energies(),
                                                 self.frequency_total_global,
                                                 'total',
                                                 None ] )
                        data_list.append( [ self.sigma_energies(),
                                                 self.frequency_null_collision,
                                                 'null collision',
                                                 None ] )
                        data_list.append( [ self.sigma_energies(),
                                                 self.frequency_total_global_max*numpy.ones(self.n_sigma),
                                                 'max',
                                                 None ] )

                        plot_curves(self.coll_freq_graph, data_list)
                        # Shut down gnuplot process, window remain visible due to persist option
                        plot_close(self.coll_freq_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.coll_freq_graph)
                        
                # Plot relative collision frequencies
                if plot_relative:
                        data_list = []
                        self.rel_coll_freq_graph = new_plot(title='Relative collision frequencies for electron scattering',
                                                            xlabel='Energy / eV',
                                                            ylabel='Relative probability / %',
                                                            logx=True,
                                                            grid=True,
                                                            style=plot_style,
                                                            persistence=XSEC_PLOT_PERSIST,
                                                            redirect_output=True)                        
                        for i in range(self.types):
                                data_list.append( [ self.sigma_energies(),
                                                         self.p_relative_elastic[i]*100,
                                                         'elastic '+self.names[i],
                                                         None ] )
                                data_list.append( [ self.sigma_energies(),
                                                         self.p_relative_ionization[i]*100,
                                                         'ionization '+self.names[i],
                                                         None ] )
                                for j in range(self.excitation_types[i]):
                                        data_list.append( [ self.sigma_energies(),
                                                                 self.p_relative_excitation[i,j]*100,
                                                                 'excitation ' + self.names[i] + ' # ' + str(j),
                                                                 None ] )                                
                                for j in range(self.dissociation_types[i]):
                                        data_list.append( [ self.sigma_energies(),
                                                                 self.p_relative_dissociation[i,j]*100,
                                                                 'dissociation ' + self.names[i] + ' # ' + str(j),
                                                                 None ] )
                        data_list.append( [ self.sigma_energies(),
                                                 self.p_relative_null_collision*100,
                                                 'null collision',
                                                 None ] )

                        plot_curves(self.rel_coll_freq_graph, data_list)
                        # Shut down gnuplot process, window remain visible due to persist option
                        plot_close(self.rel_coll_freq_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.rel_coll_freq_graph)
                        
                # Plot probability boundaries
                if plot_boundaries:
                        data_list = []
                        self.boundaries_graph = new_plot(title='Probability boundaries for electron scattering',
                                                         xlabel='Energy / eV',
                                                         ylabel='Probability limit',
                                                         logx=True,
                                                         grid=True,
                                                         style=plot_style,
                                                         persistence=XSEC_PLOT_PERSIST,
                                                         redirect_output=True)
                        for i in range(self.n_limits):
                                data_list.append( [ self.sigma_energies(),
                                                         self.probability_limits[i],
                                                         #'boundary #'+str(i),
                                                         self.probability_processes[i],
                                                         None ] )

                        plot_curves(self.boundaries_graph, data_list)
                        # Shut down gnuplot process, window remain visible due to persist option
                        plot_close(self.boundaries_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.boundaries_graph)
                        
                return status, message


        # +-------------------------------------------------------------+
        # | Methods to read cross sections for ions/neutrals scattering |
        # +-------------------------------------------------------------+

        def read_xsec_ions_elastic(self, filename, sep, ion_index, target_index, plot=False):
                """Reads the table of cross sections values for ions-neutrals elastic scattering as a function 
                   of energy from a ascii file.

                        Parameters
                        ----------

                        filename:       name of file containing cross section values at energies in the form

                                                #Comment
                                                <energy> <separator> <value>    #Comment
                                                <energy> <separator> <value>    #Comment

                                        where energies are in eV and cross section values are in m**2
                                        
                        sep:            character used to separate values (usually the tab char)
                        ion_index:      integer value indentifying the type of ion (i.e. the type of neutral that was ionized) 
                                        that elastically scatters with the neutral
                        target_index:   integer value indentifying the type of neutral target
                        plot:           if True, plot the cross section values as a function of energy,
                                        before and after interpolation

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = negative energy value found
                                         2 = negative cross-section value found
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                title    = 'Cross section for elastic scattering of ' + self.names[ion_index] + '+ on ' + self.names[target_index]
                qm_ratio = -ELECTRON_CHARGE / (self.mass[ion_index] * ATOMIC_UNIT_MASS)

                (status, message) = self._read_xsec(filename, sep, qm_ratio, target_index, ion_index, plot, title)

                if (status == 0): self.sigma_elastic_ions[ion_index, target_index] = self.sigma_return

                return (status, message)


        def read_xsec_ions_charge_exchange(self, filename, sep, ion_index, target_index, plot=False):
                """Reads the table of cross sections values for ions-neutrals charge exchange scattering 
                   as a function of energy from a ascii file.

                        Parameters
                        ----------

                        filename:       name of file containing cross section values at energies in the form

                                                #Comment
                                                <energy> <separator> <value>    #Comment
                                                <energy> <separator> <value>    #Comment

                                        where energies are in eV and cross section values are in m**2
                                        
                        sep:            character used to separate values (usually the tab char)
                        ion_index:      integer value indentifying the type of ion (i.e. the type of neutral that was ionized) 
                                        that scatters with the neutral
                        target_index:   integer value indentifying the type of neutral target


                        plot:           if True, plot the cross section values as a function of energy,
                                        before and after interpolation

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = negative energy value found
                                         2 = negative cross-section value found
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                title    = 'Cross section for charge exchange scattering of ' + self.names[ion_index] + '+ on ' + \
                           self.names[target_index] 
                qm_ratio = -ELECTRON_CHARGE / (self.mass[ion_index] * ATOMIC_UNIT_MASS)

                (status, message) = self._read_xsec(filename, sep, qm_ratio, target_index, ion_index, plot, title)

                if (status == 0): self.sigma_charge_exchange_ions[ion_index,target_index] = self.sigma_return

                return (status, message)


        def calculate_total_xsec_ions(self, debug=False):
                """Calculates main statistical parameters from cross sections for ion impact processes.

                        Calculate, for the different processes (elastic scattering, ionization, dissociation types): 
                                - total cross sections
                                - collision frequencies
                                - relative scattering probabilities
                                - probability boundaries
                        (see the module docstring for more information on these quantities and their usage)

                        Parameters
                        ----------

                        Initialized data attributes
                        ---------------------------

                        Calculated total global cross section for each ion, sum of all targets and all processes
                        self.sigma_total_global_ions:          sum of all cross section for all processes

                        Calculated cross sections for each ion/neutral combination, sum of all processes
                        self.sigma_total_ions:                 sum of all cross sections (elastic + charge exchange)
                                                                 for each ion/target combinations

                        Calculated collision frequencies for   each ion/neutral combination
                        self.frequency_elastic_ions:           collision frequencies for elastic scattering for each 
                        self.frequency_charge_exchange_ions:   collision frequencies for charge exchange sc
                        self.frequency_total_ions:             total collision frequency (elastic + charge exchange) 

                        Calculated collision frequencies for   each ion (over all neutral types)
                        self.frequency_total_global_ions:      sum of all collision frequencies, for all possible processes
                                                               and all neutral types 
                        self.frequency_total_global_max_ions:  max value of the total global frequency, over all speeds
                        self.frequency_null_collision_ions:    null collision frequency, defined as the difference between
                                                                 the maximum value (over all velocities) of the 
                                                                 total global collision frequency
                                                                 and the total global collision frequency at that speed

                        Calculated probabily limits for elastic, charge exchange and null collision
                        self-p_null_ions:                      probability to have e null collision for that ion (all targets)
                        self.probability_limits_ions:          once stated which is the target, this are the probability limits
                                                               for the possible collision types: elastic, charge exchange, null

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = error
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                self.sigma_total_global_ions            = numpy.zeros( (self.types, self.n_sigma_ions), 'd', order='F')

                self.sigma_total_ions                   = numpy.zeros( (self.types, self.types, self.n_sigma_ions), 'd', order='F')

                self.frequency_elastic_ions             = numpy.zeros( (self.types, self.types, self.n_sigma_ions), 'd', order='F')
                self.frequency_charge_exchange_ions     = numpy.zeros( (self.types, self.types, self.n_sigma_ions), 'd', order='F')
                self.frequency_total_ions               = numpy.zeros( (self.types, self.types, self.n_sigma_ions), 'd', order='F')
                
                self.frequency_total_global_ions        = numpy.zeros( (self.types, self.n_sigma_ions), 'd', order='F')
                self.frequency_total_global_max_ions    = numpy.zeros( self.types, 'd')                
                self.frequency_null_collision_ions      = numpy.zeros( (self.types, self.n_sigma_ions), 'd', order='F')

                self.p_relative_null_ions               = numpy.zeros( (self.types, self.n_sigma_ions), 'd', order='F')
                self.p_relative_elastic_ions            = numpy.zeros( (self.types, self.types, self.n_sigma_ions), 'd', order='F')
                self.p_relative_charge_exchange_ions    = numpy.zeros( (self.types, self.types, self.n_sigma_ions), 'd', order='F')

                self.probability_limits_ions            = numpy.zeros( (self.types, self.types, 2, self.n_sigma_ions), 'd', order='F')

                # Calculate total cross section (elastic + charge exchange) for each ion/target combination
                for ion in range(self.types):
                        for target in range(self.types):
                                self.sigma_total_ions[ion,target] = (   self.sigma_elastic_ions[ion,target]
                                                                      + self.sigma_charge_exchange_ions[ion,target] )

                # Calculate collision frequencies for all processes and their sum
                # f_(i,t)(v) = n_t * s_(i,t)(v) * v_rel
                for ion in range(self.types):
                        for target in range(self.types):
                                self.frequency_elastic_ions[ion,target]         = (  self.number_density[target]
                                                                                   * self.sigma_elastic_ions[ion,target]
                                                                                   * self.sigma_velocities_ions[ion] )
                                self.frequency_charge_exchange_ions[ion,target] = (  self.number_density[target]
                                                                                   * self.sigma_charge_exchange_ions[ion,target]
                                                                                   * self.sigma_velocities_ions[ion] )
                                self.frequency_total_ions[ion,target] = (   self.frequency_elastic_ions[ion,target]
                                                                          + self.frequency_charge_exchange_ions[ion,target] )
                                                               
                # Calculate total global collision frequency for each ion type, summing over all neutral types
                # f_tot_i(v) = sum{f_(i,t)(v), t}
                for ion in range(self.types):
                        for target in range(self.types):
                                self.frequency_total_global_ions[ion] = (  self.frequency_total_global_ions[ion]
                                                                         + self.frequency_total_ions[ion,target] )

                # Calculate the maximum value of the total global collision frequency over all velocities
                #   f_max = max{f_tot(v), v}
                for ion in range(self.types):
                        self.frequency_total_global_max_ions[ion] = self.frequency_total_global_ions[ion].max()

                # Calculate the null collision frequency
                #   f_null(v) = f_max - f_tot(v)
                for ion in range(self.types):
                        self.frequency_null_collision_ions[ion] = (  self.frequency_total_global_max_ions[ion] 
                                                                   - self.frequency_total_global_ions[ion]    )
                # Calculate relative scattering probabilities
                for ion in range(self.types):
                        # Probability to have a null collision (regardless of the target)
                        self.p_relative_null_ions[ion] = self.frequency_null_collision_ions[ion] / self.frequency_total_global_max_ions[ion]
                        for target in range(self.types):
                                # Relative probability to have elastic or charge exchange, for a specific target,
                                # if the collision is not null
                                self.p_relative_elastic_ions[ion, target] = (  self.frequency_elastic_ions[ion,target]
                                                                             / self.frequency_total_ions[ion,target]   )
                                self.p_relative_charge_exchange_ions[ion, target] = (  self.frequency_charge_exchange_ions[ion,target]
                                                                                     / self.frequency_total_ions[ion,target]   )
                
                # Probability boundaries to have a collision elastic, charge exchange or null
                # once the target has been identified        
                for ion in range(self.types):
                        for target in range(self.types):
                                self.probability_limits_ions[ion, target, 0] = (  (1 - self.p_relative_null_ions[ion])
                                                                                * self.p_relative_elastic_ions[ion,target]  )
                                self.probability_limits_ions[ion, target, 1] = (  self.probability_limits_ions[ion, target, 0]
                                                                                + (1 - self.p_relative_null_ions[ion])
                                                                                * self.p_relative_charge_exchange_ions[ion,target]  )
                return (status, message)


        def plot_xsec_ions(self, ion_type, plot_total=True, plot_frequencies=False, plot_relative=False, plot_boundaries=False,
                           dot_points=False):
                """Plots cross sections for ion collisions and related main statistical parameters, as a function of electron energy.

                        Plots, for the different processes (elastic scattering, ionization, dissociation types): 
                                - total cross sections
                                - collision frequencies
                                - relative probability of elastic scattering

                        Parameters
                        ----------

                        ion_type:         type of ion for which the quantities will be plotted
                        plot_total:       if True, plot the total cross section values as a function of energy
                        plot_frequencies: if True, plot the collision frequency values as a function of energy
                        plot_relative:    if True, plot the relative probability of elastic scattering
                        plot_boundaries:  if True, plot the probavility boundaries

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = error
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking

                if (not gnuplot_installed): return

                if dot_points: plot_style = 'dots'
                else:          plot_style = 'points'               

                # Plot total cross-sections                       
                if plot_total:
                        data_list = []
                        # Prepare graph
                        self.cross_section_graph = new_plot(title='Cross sections for ' + self.names[ion_type] + '+ impact',
                                                            xlabel='Energy / eV',
                                                            ylabel='Cross section / m**2',
                                                            logx=True,
                                                            logy=True,
                                                            grid=True,
                                                            style=plot_style,
                                                            persistence=XSEC_PLOT_PERSIST,
                                                            redirect_output=True)
                        # Plot total global cross section (all neutral types and all processes)
                        data_list.append( [ self.sigma_energies_ions(ion_type),
                                                 self.sigma_total_global_ions[ion_type],
                                                 'total global' + self.names[ion_type] + '+',
                                                 None ] )
                        # Plot cross sections for each neutral and process type
                        for target in range(self.types):
                                # Toal cross section (all processes)
                                data_list.append( [ self.sigma_energies_ions(ion_type),
                                                         self.sigma_total_ions[ion_type,target],
                                                         'total ' + self.names[ion_type] + '+ on ' + self.names[target],
                                                         None ] )
                                # Elastic cross section
                                data_list.append( [ self.sigma_energies_ions(ion_type),
                                                         self.sigma_elastic_ions[ion_type,target],
                                                         'elastic ' + self.names[ion_type] + '+ on ' + self.names[target],
                                                         None ] )
                                # Charge exchange cross section
                                data_list.append( [ self.sigma_energies_ions(ion_type),
                                                         self.sigma_charge_exchange_ions[ion_type,target],
                                                         'charge exchange '+self.names[ion_type]+'+ on '+ self.names[target],
                                                         None ] )
                                
                        plot_curves(self.cross_section_graph, data_list)
                        
                        plot_close(self.cross_section_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.cross_section_graph)
                                
                # Plot collision frequencies
                if plot_frequencies:
                        data_list = []
                        self.coll_freq_graph = new_plot(title='Collision frequencies for ' + self.names[ion_type] + '+ impact',
                                                        xlabel='Energy / eV',
                                                        ylabel='Frequency / Hz',
                                                        logx=True,
                                                        logy=True,
                                                        format_x='%g',
                                                        format_y='%g',
                                                        grid=True,
                                                        style=plot_style,
                                                        persistence=XSEC_PLOT_PERSIST,
                                                        redirect_output=True)
                        
                        for target in range(self.types):
                                data_list.append( [ self.sigma_energies_ions(ion_type),
                                                    self.frequency_elastic_ions[ion_type,target],
                                                    'elastic ' + self.names[ion_type] + '+ on ' + self.names[target],
                                                    None ] )
                                data_list.append( [ self.sigma_energies_ions(ion_type),
                                                    self.frequency_charge_exchange_ions[ion_type,target],
                                                    'charge exchange ' + self.names[ion_type] + '+ on ' + self.names[target],
                                                    None ] )
                                data_list.append( [ self.sigma_energies_ions(ion_type),
                                                    self.frequency_total_ions[ion_type, target],
                                                    'total ' + self.names[ion_type] + '+ on ' + self.names[target],
                                                    None ] )
                        data_list.append( [ self.sigma_energies_ions(ion_type),
                                            self.frequency_total_global_ions[ion_type],
                                            'total global ' + self.names[ion_type] + '+',
                                            None ] )                                
                        data_list.append( [ self.sigma_energies_ions(ion_type),
                                            self.frequency_null_collision_ions[ion_type],
                                            'null collision ' + self.names[ion_type] + '+', None ] )
                        data_list.append( [ self.sigma_energies_ions(ion_type),
                                            self.frequency_total_global_max_ions[ion_type]*numpy.ones(self.n_sigma_ions),
                                            'max ' + self.names[ion_type] + '+', None ] )

                        plot_curves(self.coll_freq_graph, data_list)                        

                        # Shut down gnuplot process, window remain visible due to persist option
                        plot_close(self.coll_freq_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.coll_freq_graph)

                # Plot relative collision frequencies
                if plot_relative:
                        data_list = []
                        self.cross_sect_rel_graph = new_plot(title='Relative probability of ion scattering for '
                                                             + self.names[ion_type]
                                                             + '+ impact',
                                                             xlabel='Energy / eV',
                                                             ylabel='Relative probability / %',
                                                             logx=True,
                                                             grid=True,
                                                             ymin = 0,
                                                             ymax = 100,
                                                             style=plot_style,
                                                             persistence=XSEC_PLOT_PERSIST,
                                                             redirect_output=True)
                        data_list.append( [ self.sigma_energies_ions(ion_type),
                                            self.p_relative_null_ions[ion_type]*100,
                                            'null '+self.names[ion_type]+'+',
                                            None ] )                        
                        for target in range(self.types):
                                data_list.append( [ self.sigma_energies_ions(ion_type),
                                                    self.p_relative_elastic_ions[ion_type,target]*100,
                                                    'elastic '+self.names[ion_type]+'+ on '+self.names[target],
                                                    None ] )
                                data_list.append( [ self.sigma_energies_ions(ion_type),
                                                    self.p_relative_charge_exchange_ions[ion_type,target]*100,
                                                    'charge exchange '+self.names[ion_type]+'+ on '+self.names[target],
                                                    None ] )                                

                        plot_curves(self.cross_sect_rel_graph, data_list)
                        
                        # Shut down gnuplot process, window remain visible due to persist option
                        plot_close(self.cross_sect_rel_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.cross_sect_rel_graph)

                # Plot probability boundaries
                if plot_boundaries:
                        for target in range(self.types):                        
                                data_list = []
                                self.cross_sect_lim_graph = new_plot(title='Probability boundaries for '
                                                                     + self.names[ion_type]
                                                                     + '+ impact on ' + self.names[target],
                                                                     xlabel='Energy / eV',
                                                                     ylabel='Probability limit',
                                                                     logx=True,
                                                                     grid=True,
                                                                     ymin = 0,
                                                                     ymax = 1,
                                                                     style=plot_style,
                                                                     persistence=XSEC_PLOT_PERSIST,
                                                                     redirect_output=True)

                                data_list.append( [ self.sigma_energies_ions(ion_type),
                                                    self.probability_limits_ions[ion_type,target,0],
                                                    'limit 0 (elastic) '+self.names[ion_type]+'+ on '+self.names[target],
                                                    None ] )
                                data_list.append( [ self.sigma_energies_ions(ion_type),
                                                    self.probability_limits_ions[ion_type,target,1],
                                                    'limit 1 (charge exchange) '+self.names[ion_type]+'+ on '+self.names[target],
                                                    None ] )                                

                                plot_curves(self.cross_sect_lim_graph, data_list)
                        
                                # Shut down gnuplot process, window remain visible due to persist option
                                plot_close(self.cross_sect_lim_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                                del(self.cross_sect_lim_graph)                        

                return status, message


        # +----------------------------------------------------------------------+
        # | Methods to read cross sections for electrons recombination with ions |
        # +----------------------------------------------------------------------+

        def read_xsec_ele_ion_recomb(self, filename, sep, ion_index, rec_type=None, diss_type=None, plot=False):
                """Reads the table of cross section values for recombination of electrons with ions.

                        Parameters
                        ----------

                        filename:       name of file containing cross section values at energies in the form

                                                #Comment
                                                <energy> <separator> <value>    #Comment
                                                <energy> <separator> <value>    #Comment

                                        where energies are in eV and cross section values are in m**2
                                        
                        sep:            character used to separate values (usually the tab character)
                        ion_index:      integer value indentifying the type of ion (must be a molecule) 
                                        with which the elctron recombine
                        rec_type:       type of recombination process
                                        'dissociative' -> dissociative recombination of e- with a molecular ion
                                        '3body'        -> recombination of e- with ion, with transfer of the released energy 
                                                          to another e-
                                        'radiative'    -> recombination of e- with ion, with radiative release of excess energy

                        diss_type:      integer value indentifying the type of dissociative recombination process to which
                                        the cross section is referred (only for dissociative recombination processes)
                                        (the number of dissociation types for each molecule type is defined 
                                        in the "neutrals.csv" file)
                        plot:           if True, plot the cross section values as a function of energy,
                                        before and after interpolation


                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = negative energy value found
                                         2 = negative cross-section value found
                                         3 = ion index is outside allowed bounds
                                         4 = unspecified recombination type
                                         5 = unknown recombination type
                                         6 = dissociation process index is missing
                                         7 = dissociation process index is out of bounds
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                if ( (ion_index < 0) or (ion_index > self.types-1)):
                        status = 3
                        message = 'ion type ' + str(ion_index) + ' is out of bounds'
                        return(status, message)
                if (rec_type is None):
                        status = 4
                        message = 'unspecified recombination type'
                        return(status, message)
                if (rec_type not in ('dissociative', '3body', 'radiative')):
                        status = 5
                        message = 'unknown recombination type \"' + str(rec_type) + '\"'
                        return(status, message)
                if (rec_type == 'dissociative'):
                        if (diss_type is None):
                                status = 6
                                message = 'dissociation process index is missing'
                                return (status, message)
                        if (diss_type > self.dissociation_types[ion_index]):
                                status = 7
                                message = 'dissociation process index ' + str(diss_type) + ' is out of bounds'
                                return (status, message)

                title    = 'Cross section for ' + rec_type + ' recombination of e- with ' + self.names[ion_index]
                if (rec_type =='dissociative'): 
                        title = title + ' (process #' + str(diss_type) + ')'
                qm_ratio = - ELECTRON_CHARGE / ELECTRON_MASS

                (status, message) = self._read_xsec(filename, sep, qm_ratio, ion_index, None, plot, title)

                if (status == 0): 
                        if (rec_type == 'dissociative'): self.sigma_recombination_diss[ion_index,diss_type] = self.sigma_return
                        if (rec_type == '3body'):        self.sigma_recombination_3body[ion_index]          = self.sigma_return
                        if (rec_type == 'radiative'):    self.sigma_recombination_rad[ion_index]            = self.sigma_return

                return (status, message)


        def calculate_total_xsec_ele_ion_recomb(self):
                """Calculates main statistical parameters from cross sections.

                        Calculate, for the different processes of electron-ion recombination:
                                - total cross sections
                                - recombination rate coefficients (non averaged), i.e. product cross section by electron velocity
                        (see the module docstring for more information on these quantities and their usage)

                        When calculating the recombination process for an electron with a given ion type, 
                        the background density involved in the calculation is the density of that ion type.
                        Since it is not constant [*], it is not possible to pre-calculate 
                        recombination frequencies, nor recombination rates. They must be calculated during simulation
                        at each timestep based on the ion density at that time.

                        [*] Instead, neutral density can be considered appoximately constant, if the ionization degree is very low
                        so collision frequencies with neutrals can be pre-calculated.

                        Parameters
                        ----------

                        Initialized data attributes
                        ---------------------------

                        self.ratecoeff_recomb_diss:     rate coefficients (sigma*v) for dissociative recombination processes 
                                                             of each ion type

                        self.sigma_total_recomb_diss:   sum of dissociative recombination cross sections for all neutral types
                                                             and all dissociation processes for each neutral type

                        self.sigma_total_global_recomb: sum of all cross section for all recombination processes

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = error
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking      

                self.sigma_total_recombination_diss    = numpy.zeros( self.n_sigma, 'd' )
#               self.sigma_total_recombination_3body   = numpy.zeros( self.n_sigma, 'd' )
#               self.sigma_total_recombination_rad     = numpy.zeros( self.n_sigma, 'd' )
                self.sigma_total_global_recombination  = numpy.zeros( self.n_sigma, 'd' )
                self.ratecoeff_recomb_diss             = numpy.zeros( (self.types, MAX_DISSOCIATION_TYPES, self.n_sigma), 
                                                                      'd', order = 'F')
                # Calculate total cross sections for each recombination process
                # (at the moment only dissociative recombination is considered)
                for ion_index in range(self.types):
                        for diss_type in range(self.dissociation_types[ion_index]):
                                self.sigma_total_recombination_diss = self.sigma_total_recombination_diss + \
                                                                      self.sigma_recombination_diss[ion_index,diss_type]

                # Calculate total global cross-section, for all the possible recombination types
                # (at the moment only dissociative recombination is considered,
                # so it's equal to total cross section for that type of recombination)
                self.sigma_total_global_recombination  = self.sigma_total_recombination_diss
#                                                       + self.sigma_total_recombination_3body
#                                                       + self.sigma_total_recombination_rad

                # Calculate rate coefficients (not averaged) for each recombination process
                # (at the moment only dissociative recombination is considered)
                # rate coefficient is the product cross section by velocity -> sigma(v) * v
                for ion_index in range(self.types):
                        for diss_type in range(self.dissociation_types[ion_index]):
                                self.ratecoeff_recomb_diss[ion_index,diss_type]= self.sigma_recombination_diss[ion_index,diss_type]*\
                                                                                 self.sigma_velocities
#                        self.ratecoeff_recomb_3body[ion_index] = self.sigma_recombination_3body[ion_index] * self.sigma_velocities     
#                        self.ratecoeff_recomb_rad[ion_index]   = self.sigma_recombination_rad[ion_index]   * self.sigma_velocities

                return (status, message)


        def plot_xsec_ele_ion_recomb(self, plot_total=False, plot_rates=False, dot_points=False):
                """Plots cross sections for electrons collisions and related main statistical parameters, 
                   as a function of electron energy.

                        Plots, for the different processes (elastic scattering, ionization, dissociation types): 
                                - total cross sections
                                - collision rates 

                        Parameters
                        ----------

                        plot_total:       if True, plot the total cross section values as a function of energy
                        plot_rates:       if True, plot the collision rate values as a function of energy

                        Returns
                        -------

                        (status, message)
                                status:  0 = no error
                                         1 = error
                                message: a string containing an error message or 'Ok'
                """

                status, message  = 0, 'OK'              # This variable is used for error checking

                if (not gnuplot_installed): return

                if dot_points: plot_style = 'dots'
                else:          plot_style = 'points'                

                # Plot total cross-sections
                if plot_total:
                        data_list = []
                        self.el_ion_recomb_graph = new_plot(title='Cross sections for e-/ion recombination',
                                                            xlabel='Energy / eV',
                                                            ylabel='Cross section / m**2',
                                                            logx=True,
                                                            logy=True,
                                                            format_x='%g',
                                                            format_y='%g',
                                                            grid=True,
                                                            style=plot_style,
                                                            persistence=XSEC_PLOT_PERSIST,
                                                            redirect_output=True)                                                      
                        data_list.append( [ self.sigma_energies(),
                                                 self.sigma_total_global_recombination, 
                                                 'total global',
                                                 None ] )

                        # Plot total cross sections for each process type
                        data_list.append( [ self.sigma_energies(),
                                                 self.sigma_total_recombination_diss, 
                                                 'total dissociative recombination',
                                                 None ] )
                        # Plot cross sections for each process type and gas type
                        for i in range(self.types):
                                for j in range(self.dissociation_types[i]):
                                        data_list.append( [ self.sigma_energies(),
                                                                 self.sigma_recombination_diss[i,j],
                                                                 'dissociative recombination ' + self.names[i] + ' # '+str(j),
                                                                 None ] )
                                        
                        plot_curves(self.el_ion_recomb_graph, data_list)
                                         
                        plot_close(self.el_ion_recomb_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.el_ion_recomb_graph)

                # Plot collision rates
                if plot_rates:
                        data_list = []
                        self.coll_rates_graph = new_plot(title='Collision rate coefficients for e-/ion recombination',
                                                         xlabel='Energy / eV',
                                                         ylabel='Collision rate coefficients / m**3 s**-1',
                                                         logx=True,
                                                         logy=True,
                                                         format_x='%g',
                                                         format_y='%g',
                                                         grid=True,
                                                         style=plot_style,
                                                         persistence=XSEC_PLOT_PERSIST,
                                                         redirect_output=True)                                                       
                        for i in range(self.types):
                                for j in range(self.dissociation_types[i]):
                                        self.plot_lsit.append( [ self.sigma_energies(),
                                                                 self.ratecoeff_recomb_diss[i,j],
                                                                 ( 'dissociative recombination '
                                                                   + self.names[i]
                                                                   + ' # '
                                                                   + str(j) ),
                                                                 None ] )

                        plot_curves(self.coll_rates_graph, data_list)

                        plot_close(self.coll_rates_graph, keep_output=True, delay=DEL_DATA_FILES_DELAY)
                        del(self.coll_rates_grap)

                return status, message


        def initialize_savefile(self, filename_neutral, append=False, sep='\t', ext='.csv'):
                """Initializes the file to which simulation data about neutral particles will be saved

                   Parameters
                   ----------
                   filename_neutral:      name of the file to which the simulation quantities will be saved
                   sep:                   character or sequence to separate data columns
                   ext:                   extension to give to filename (deafult is '.csv')

                   Initialized data attributes
                   ---------------------------

                   append:                if True, open the file for appendig new data to the end
                                          header will not be re-written
                   self.sep:              character or sequence to separate data columns
                   self.f_neutral:        file to which values of several quantities will be saved
                """

                self.sep = sep
                self.filename = str(filename_neutral) + ext
                
                # Open data file where to save time evolution of mean values
                if append:
                        data_file = open(self.filename,'a')
                else:
                        data_file = open(self.filename,'w')
                        # Write header
                        data_file.write('\"t[ns]\"' + self.sep )                
                        for i in range(self.types):
                                for j in range(self.dissociation_types[i]):
                                        data_file.write('\"Diss_rate ' + self.names[i] + '#'+str(j) + \
                                                          ' [m**-3*s**-1]\"' + self.sep   )
                                        data_file.write('\"Diss_rate const ' + self.names[i] + '#'+str(j) + \
                                                          ' [m**3*s**-1]\"' + self.sep   )
                        data_file.write(EOL)
                data_file.close()


        def save_data_to_files(self, time):
                """Saves actual data values of dissociation rates to files """

                data_file = open(self.filename,'a')
                data_file.write( str(1E9*time) + self.sep )                
                for i in range(self.types):
                        for j in range(self.dissociation_types[i]):
                                data_file.write( str(self.dissociation_rate[i][j]) + self.sep )
                                data_file.write( str(self.dissociation_rate_constant[i][j]) + self.sep )
                data_file.write(EOL)
                data_file.close()



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Simulation of a capacitively coupled plasma discharge. """

# +--------------------------------+
# | Import required Python modules |
# +--------------------------------+

# Mudules from the standard Python library
import sys
import math
from random import random
from optparse import OptionParser
import tkinter as tk
#import tkMessageBox

# Modules provided by the Python community
import numpy

# Import required constants and parameters
from pysica.parameters import *
from pysica.constants import *
from pysica.plasmapro.ccpla_defaults import *

# Import required modules, classes, and functions
from pysica.io.io_screen import wait_input, clear_screen
from pysica.managers import unit_manager
from pysica.plasmapro import ccpla_gui
from pysica.plasmapro.discharge import reactors, target_particles, moving_particles, particle_mover
from pysica.plasmapro.ccpla_init import *
from pysica.plasmapro.ccpla_print import *

# The gnuplot_installed flag is set to False if gnuplot is not installed
from pysica.managers.gnuplot_manager import gnuplot_installed


# +---------------------------------------------------+
# | Manage how to write messages and exit the program |
# +---------------------------------------------------+

class CcplaMessageWindow(tk.Frame):
    def __init__(self, message, error, master=None):
        tk.Frame.__init__(self, master)

        if error:
            self.back         = 'light yellow'
            self.fore         = 'red'
            self.button_label = 'Exit'
        else:
            self.back         = 'light grey'
            self.fore         = 'blue'
            self.button_label = 'Dismiss'
        self.message = tk.Message(self,
                                  text=message,
                                  background=self.back,
                                  foreground=self.fore,
                                  width=1500,
                                  justify=tk.CENTER,
                                  relief=tk.RIDGE)
        self.message.grid()
        self.button  = tk.Button(self, text=self.button_label, command=self.master.destroy)
        self.button.grid()
        self.grid()
 
def create_message_window(message, error):
    root = tk.Tk()
    if error: root.title("Ccpla Error")
    else:     root.title("Ccpla Message")
    root.resizable(False, False)

    message_window = CcplaMessageWindow(message, error, master=root)

    root.mainloop()

def exit_ccpla(message="Exiting program", gui=False, error=False):
    if not gui:
        sys.exit(EOL + message + EOL)
    else:
        create_message_window(message, error)
        sys.exit()

def print_message(message, gui=False, error=False):
    if gui:
        create_message_window(message, error)
    else:
        if error: message = 'ERROR: ' + message
        print('\n' + message + '\n')
             
# +------------------------+
# | Command line arguments |
# +------------------------+

usage = "usage: %prog [options]"        # String to be given as program usage description
parser = OptionParser(usage)

# Do not start simulation, only print parameters values and exit
help_string = "show parameters values and exit"
parser.add_option("-p", "--print-only", action="store_true",  dest="print_only", default=False, help=help_string)

# Only save default values 
help_string = "write default parameters to file \"" + FILENAME_DEFAULTS + "\" and exit"
parser.add_option("-s", "--save-defaults", action="store_true",  dest="save_defaults", default=False, 
                  help=help_string)

# Batch mode
help_string = "suppress all input from user"
parser.add_option("-b", "--batch-mode", action="store_true", dest="batch_mode", default=batch_mode,
                  help=help_string)

# GUI mode
help_string = "start GUI"
parser.add_option("-g", "--gui-mode", action="store_true", dest="gui_mode", default=gui_mode,
                  help=help_string)

# Width of text window in GUI mode
help_string = ("Width of text window in GUI ["
               + str(TEXT_WINDOW_WIDTH_MIN) + ".." + str(TEXT_WINDOW_WIDTH_MAX)
               + "] (default=" + str(text_window_width) + ")")
parser.add_option("-W", "--text-window-width", action="store", type="int", dest="text_window_width",
                  default=text_window_width, help=help_string)

# Height of text window in GUI mode
help_string = ("Height of text window in GUI ["
               + str(TEXT_WINDOW_HEIGHT_MIN) + ".." + str(TEXT_WINDOW_HEIGHT_MAX)
               + "] (default=" + str(text_window_height) + ")")
parser.add_option("-H", "--text-window-height", action="store", type="int", dest="text_window_height",
                  default=text_window_height, help=help_string)

# Font dimension in text window in GUI mode
help_string = ("Font size in GUI text window ["
               + str(TEXT_WINDOW_FONT_SIZE_MIN) + ".." + str(TEXT_WINDOW_FONT_SIZE_MAX)
               + "] (default=" + str(text_window_font_size) + ")")
parser.add_option("-F", "--text-window-font", action="store", type="int", dest="text_window_font_size",
                  default=text_window_font_size, help=help_string)

# Redirect output
help_string = "redirect screen output to file \'" + FILENAME_OUTPUT_LOG + "\'"
parser.add_option("-o", "--redirect-output", action="store_true", dest="redirect_output", default=redirect_output,
                  help=help_string)        

# Redirect error messages
help_string = "redirect error messages to file \'" + FILENAME_ERROR_LOG + "\'"
parser.add_option("-e", "--redirect-errors", action="store_true", dest="redirect_error_messages",
                  default=redirect_error_messages,
                  help=help_string)

# Verbosity level
help_string = "verbosity level [0..3] (default="+str(verbosity)+")"
parser.add_option("-v", "--verbosity", action="store", type="int", dest="verbosity",  default=verbosity,
                  help=help_string)

# Debug level for python code
help_string = "Python debug level [0..2] (default="+str(debug_level_python)+")"
parser.add_option("-d", "--debug-level-python", action="store", type="int", dest="debug_lev",
                  default=debug_level_python, help=help_string)

# Debug level for Fotran code
help_string = "Fortran debug level [0..3] (default="+str(debug_level_fortran)+")"
parser.add_option("-D", "--debug-level-fortran", action="store", type="int", dest="debug_lev_for",
                  default=debug_level_fortran, help=help_string)        
        
# Plot cross sections before starting simulation
help_string = "plot cross sections graphs before start"
parser.add_option("-x", "--graph-xsec", action="store_true",  dest="plot_xsec", default=False, help=help_string)

(cl_options, args) = parser.parse_args()

if cl_options.redirect_error_messages:
    sys.stderr = open(FILENAME_ERROR_LOG, 'w')
    
if cl_options.redirect_output:
    sys.stdout = open(FILENAME_OUTPUT_LOG, 'w')
    

# +-----------------------------+   
# | Error checking and warnings |
# +-----------------------------+

if (cl_options.verbosity not in list(range(4))):
    exit_ccpla('ERROR: verbosity must be in range 0..3' + EOL,
               gui=cl_options.gui_mode, error=True)
    
if (cl_options.debug_lev not in list(range(3))):
    exit_ccpla('ERROR: Python debug level must be in range 0..2' + EOL,
               gui=cl_options.gui_mode, error=True)
    
if (cl_options.debug_lev_for not in list(range(4))):
    exit_ccpla('ERROR: Fortran debug level must be in range 0..3' + EOL,
               gui=cl_options.gui_mode, error=True)
    
if cl_options.gui_mode:
    if not gnuplot_installed:        
        exit_ccpla('ERROR: gnuplot is not installed on your system\n'
                   + 'GUI mode is not available', gui=True)            
    cl_options.verbosity = 0
    cl_options.batch_mode = True
    
if ( (cl_options.text_window_width < TEXT_WINDOW_WIDTH_MIN)
     or (cl_options.text_window_width > TEXT_WINDOW_WIDTH_MAX) ):
    exit_ccpla('ERROR: text window width must be in range '
               + str(TEXT_WINDOW_WIDTH_MIN)
               + '..'
               + str(TEXT_WINDOW_WIDTH_MAX)
               + EOL,
               gui=cl_options.gui_mode,
               error=True)
    
if ( (cl_options.text_window_height < TEXT_WINDOW_HEIGHT_MIN)
     or (cl_options.text_window_height > TEXT_WINDOW_HEIGHT_MAX) ):
    exit_ccpla('ERROR: text window height must be in range '
               + str(TEXT_WINDOW_HEIGHT_MIN)
               + '..'
               + str(TEXT_WINDOW_HEIGHT_MAX)
               + EOL,
               gui=cl_options.gui_mode,
               error=True)
    
if ( (cl_options.text_window_font_size < TEXT_WINDOW_FONT_SIZE_MIN)
     or (cl_options.text_window_font_size > TEXT_WINDOW_FONT_SIZE_MAX) ):
    exit_ccpla('ERROR: text window font size must be in range '
               + str(TEXT_WINDOW_FONT_SIZE_MIN)
               + '..'
               + str(TEXT_WINDOW_FONT_SIZE_MAX)
               + EOL,
               gui=cl_options.gui_mode,
               error=True)

if ((cl_options.print_only) and (cl_options.gui_mode)):
    exit_ccpla('ERROR: -p option is not compatible with GUI mode', gui=True, error=True)
    
if (cl_options.plot_xsec):
    if not gnuplot_installed:
        exit_ccpla('ERROR: gnuplot is not installed on your system\n'
                   + 'plotting of cross section is not possible',
                   gui=cl_options.gui_mode,
                   error=True)
    elif cl_options.gui_mode:        
        cl_options.plot_xsec = False
        print_message(message='WARNING: -x option is disabled in GUI mode\n'
                      + 'cross section graphs can be accessed from the \"Parameters\" menu entry', 
                      gui=True,
                      error=False)        
    elif cl_options.batch_mode:
        cl_options.plot_xsec = False        
        print_message(message='WARNING: plotting of cross sections is disabled in batch mode', 
                      gui=False,
                      error=False)

cl_options.text_window_font = (TEXT_WINDOW_FONT_TYPE, str(cl_options.text_window_font_size))

# +----------------------+
# | Print copyright info |
# +----------------------+

if (not cl_options.gui_mode): print(GPL_MESSAGE)

if (not cl_options.batch_mode): wait_input()
    

# +-------------------------+
# | Read configuration file |
# +-------------------------+

parameters        = ConfigurationOptions()
(status, message) = initialize_parameters(parameters,
                                          verbose=(cl_options.verbosity > 0),
                                          saveonly=cl_options.save_defaults,
                                          filename_config=FILENAME_CONFIG,
                                          filename_defaults=FILENAME_DEFAULTS)
if (status != 0):
        exit_ccpla(EOL + message + EOL, gui=cl_options.gui_mode, error=True)
if (cl_options.save_defaults):
        exit_ccpla('\nDefault values of parameters were saved to file: \"' + FILENAME_DEFAULTS + '\"\n',
                   gui=cl_options.gui_mode, error=False)

# +------------------------------------+
# | Initialize reactor properties data |
# +------------------------------------+

ccp = reactors.CcpProperties(parameters.distance,
                             parameters.length,
                             parameters.V_bias,
                             parameters.frequency,
                             parameters.phase,
                             parameters.N_cells,
                             parameters.lateral_loss)


# +---------------------+
# | Load gas properties |
# +---------------------+

if (cl_options.verbosity > 0): print('\nCreating neutrals ensamble from file \"' + FILENAME_NEUTRALS + '\" ...')
neutrals = target_particles.TargetParticles( parameters.N_sigma,
                                             parameters.N_sigma_ions,
                                             parameters.T_neutrals,
                                             parameters.p_neutrals,
                                             parameters.min_scattered,
                                             parameters.isactive_recomb,
                                             filename=FILENAME_NEUTRALS )
(status, message) = neutrals.read_error
if (status != 0):
    ERROR = '\nERROR in file \"'+FILENAME_NEUTRALS+'\": '
    exit_ccpla(ERROR + message + EOL, gui=cl_options.gui_mode, error=True)

if (cl_options.verbosity > 0): print('\nReading gases properties from file \"' + FILENAME_NEUTRALS + '\" ...')    
(status, message) = neutrals.read_properties( FILENAME_NEUTRALS, '\t', 
                                              parameters.e_min_sigma,
                                              parameters.e_max_sigma,
                                              parameters.e_min_sigma_ions,
                                              parameters.e_max_sigma_ions, 
                                              debug=(cl_options.debug_lev>1) )
if (status != 0):
    ERROR = '\nERROR in file \"'+FILENAME_NEUTRALS+'\": '
    exit_ccpla(ERROR + message + EOL, gui=cl_options.gui_mode, error=True)

#neutrals.min_scattered = parameters.min_scattered


# +---------------------+
# | Load cross sections |
# +---------------------+

(status, message) = initialize_cross_sections(neutrals, cl_options, parameters.isactive_recomb)
if (status != 0): exit_ccpla(message + EOL, gui=cl_options.gui_mode, error=True)


# +-----------------------------------+
# | Electrons and ions initialization |
# +-----------------------------------+

# Define the ensamble of charged particles moving in the plasma: electrons and ions
if (cl_options.verbosity > 0): print('\nDefining charged particles ensambles ...')
charges = moving_particles.MovingParticles(neutrals.types+1, parameters.Nmax_particles, parameters.start_weight, parameters.rescale_factor)

(status, message) = initialize_ensambles(charges, neutrals, parameters, cl_options)
if (status!=0): exit_ccpla(message + EOL, gui=cl_options.gui_mode, error=True)


# +-----------------------------------+
# | If it was requested, call the GUI |
# +-----------------------------------+

if cl_options.gui_mode:
    message = ccpla_gui.main(charges, neutrals, ccp, parameters, cl_options)
    exit_ccpla()


# +---------------------------------------+
# | Initialization of save directory name |
# +---------------------------------------+

if (parameters.save_delay > 0):
    initialize_filenames(neutrals, parameters)
    create_save_dir(parameters)

       
# +-------------------------------------------------+
# | Print some information on simulation parameters |
# +-------------------------------------------------+

if (cl_options.verbosity > 0):
    print()
    print(print_simulation_information(neutrals, ccp, parameters, cl_options))
    print()
    if (not cl_options.batch_mode): wait_input()        
    print(print_gas_information(neutrals))
    print() 
    if (not cl_options.batch_mode): wait_input()

# If requested by command-line option, exit the script after printing information
if cl_options.print_only: exit_ccpla()


# +--------------------------+
# | Data file initialization |
# +--------------------------+

if (parameters.save_delay > 0):
    charges.initialize_savefiles(parameters.filename_stat_ele,
                                 parameters.filename_distrib_ele,
                                 parameters.filename_distrib_ion,
                                 append=False, sep='\t', ext=EXT)
    neutrals.initialize_savefile(parameters.filename_stat_neu, append=False, sep='\t', ext=EXT)
    ccp.initialize_savefiles(parameters.filename_I, parameters.filename_V, append=False, sep='\t', ext=EXT)    

        
# +----------------+
# | Main iteration |
# +----------------+

if (cl_options.verbosity > 0): print('\nSTARTING SIMULATION')

# Initialize time and main loop index
#charges.time = 0
#charges.dt   = parameters.dt

# Data save counter
i_save_data   = 1

while (charges.n_active(0) > 0):

    if (cl_options.debug_lev > 0): 
        print('t= ' + unit_manager.print_unit(charges.time,'s') + ' V= ' + unit_manager.print_unit(ccp.V,'V',3))
        print('E      min, max, mean, sigma [eV] ', charges.e_min(0), charges.e_max(0), 
                                                    charges.e_average(0), charges.e_sigma(0))
        print('theta  min, max, mean, sigma [deg]', math.degrees(charges.theta_min(0)), 
                                                    math.degrees(charges.theta_max(0)), 
                                                    math.degrees(charges.theta_average(0)),
                                                    math.degrees(charges.theta_sigma(0))) 
        print()

    if (cl_options.verbosity > 2):
        time_before          = charges.time
        n_active_el_before   = charges.n_active(0)
        electric_bias_before = ccp.V

    particle_mover.move_particles(charges, neutrals, ccp, parameters, cl_options)

    if (cl_options.verbosity > 2):
        clear_screen()
        print(print_runtime_info(charges,
                                 neutrals,
                                 ccp,
                                 parameters,
                                 cl_options,
                                 time_before,
                                 electric_bias_before,
                                 n_active_el_before))
        
    # Save data to files
    if (parameters.save_delay > 0):
        if ( (i_save_data >= parameters.save_delay) and (charges.n_active(0) > 0) ):
            charges.save_data_to_files()
            neutrals.save_data_to_files(charges.time)
            ccp.save_data_to_files(charges.time)
            i_save_data = 0
        i_save_data += 1

    # Exit the loop if a maximum duration was provided and it has been reached
    if ( (parameters.sim_duration > 0) and (charges.time >= parameters.sim_duration) ): break
        
if (cl_options.verbosity > 0): 
    if (charges.time < parameters.sim_duration):
        print('\n\nSIMULATION INTERRUPTED (NO MORE ELECTRONS)\n\n')
    else:
        print('\n\nSIMULATION COMPLETED\n\n')



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Simulation of a capacitively coupled plasma discharge. """

import math
from numpy import isnan, isinf

from pysica.managers import unit_manager
from pysica.parameters import *
from pysica.constants import *
from pysica.plasmapro.ccpla_defaults import *

# +-----------+
# | Functions |
# + ----------+

def print_simulation_information(neutrals, ccp, parameters, options,
                                 print_physical=True, print_simulation=True,
                                 print_output=True, print_filenames=True):
    string = ''
    # Print information about physical parameters
    if print_physical:
        string += 'PHYSICAL PARAMETERS\n'
        string +=('Total pressure of the gas mixture:   ' + unit_manager.print_unit(parameters.p_neutrals,'Pa',4)
                                                          + ' ('
                                                          + unit_manager.print_unit(parameters.p_neutrals_torr,'Torr',4)
                                                          + ')' + '\n')
        string +=('Temperature of the gas mixture:      ' + unit_manager.print_unit(parameters.T_neutrals, 'K',4)
                                                          + ' ('
                                                          + unit_manager.print_unit(parameters.T_neutrals-ZERO_CELSIUS,'C',4)
                                                          + ')' + '\n')
        string += 'Total number density of molecules:   '  + unit_manager.print_exp(parameters.neutrals_density, 2) + ' m**-3' + '\n'
        string += 'Distance between electrodes:         '  + unit_manager.print_unit(parameters.distance,'m') + '\n'
        string += 'Length of electrodes:                '  + unit_manager.print_unit(parameters.length,'m') + '\n'
        string += 'Area of electrodes                   '  + unit_manager.print_exp(ccp.area, 2) + ' m**2' + '\n'
        string += 'Plasma volume:                       '  + unit_manager.print_exp(ccp.volume, 2) + ' m**3' + '\n'
        string += 'Starting electrons number density:   '  + unit_manager.print_exp(parameters.start_e_density, 2) + ' m**-3' + '\n'
        string += 'Starting ionization degree:          '  + unit_manager.print_exp(parameters.start_ion_deg, 2) + '\n'
        string += 'Electric bias peak value:            '  + unit_manager.print_unit(parameters.V_bias,'V', 4)  + '\n'
        string += 'Mean electric field intensity:       '  + unit_manager.print_unit(ccp.E_peak,'V/m', 4) + '\n'
        string += 'Electric bias frequency:             ' 
        if (ccp.frequency == 0.0):
            string += 'STATIC' + '\n'
        else:
            string += unit_manager.print_unit(ccp.frequency,'Hz', 4) + '\n'
            string += 'Electric bias period:                ' + unit_manager.print_unit(ccp.period,'s', 4) + '\n'
            string += 'Electric bias phase at t=0:          ' + str(parameters.phase * 180.0 / math.pi) + ' deg' + '\n'

    # Print information about simulation parameters
    if print_simulation:
        if not options.gui_mode: string += '\n'
        string += 'SIMULATION PARAMETERS\n'
        string += 'Maximum  number of particles:        ' + str(parameters.Nmax_particles) + '\n'
        string += 'Starting number of electrons:        ' + str(parameters.N0_electrons) + '\n'
        string += 'Starting computational weight:       ' + unit_manager.print_exp(parameters.start_weight,3) + '\n'
        string += 'Rescaling factor                     ' + str(parameters.rescale_factor) + '\n'
        string += 'Number of cells in PIC scheme:       ' + str(parameters.N_cells) + '\n'
        string += 'Cell dimension:                      ' + unit_manager.print_unit(ccp.delta_grid,'m', 4) + '\n'
        string += 'Collisions required for fast method: ' + str(neutrals.min_scattered) + '\n'
        string += 'Simulation timestep:                 '
        if parameters.dt_var:
            string += 'variable' + '\n'
            string += 'Max allowed collision frequency:     ' + str(parameters.maxcollfreq*100) + ' %' + '\n'
        else:
            string += unit_manager.print_unit(parameters.dt,'s') + '\n'
        string += 'Data output interval:                ' + unit_manager.print_unit(parameters.dt_output,'s', 4) + '\n'
        string += 'Required simulation time:            '
        if (parameters.sim_duration > 0):
            string += unit_manager.print_unit(parameters.sim_duration,'s', 4)  + '\n'
        else:
            string += 'unlimited\n'
        string += 'Electron and ion lateral loss:       '
        if ccp.lateral_loss:
            string += 'ON' + '\n'
        else:
            string += 'OFF' + '\n'
        string += 'Electron-ion recombination:          ' 
        if neutrals.isactive_recomb:
            string += 'ON' + '\n'
        else:
            string += 'OFF' + '\n'
        string += 'Electron impact cross section        ' + '\n'
        string += '- Number of cross section values:    ' + str(parameters.N_sigma) + '\n'
        string += '- Minimum cross section energy:      ' + unit_manager.print_unit(parameters.e_min_sigma, 'eV', 4) + '\n'
        string += '- Maximum cross section energy:      ' + unit_manager.print_unit(parameters.e_max_sigma, 'eV', 4) + '\n'
        string += 'Ion impact cross section             ' + '\n'
        string += '- Number of cross section values:    ' + str(parameters.N_sigma_ions) + '\n'
        string += '- Minimum cross section energy:      ' + unit_manager.print_unit(parameters.e_min_sigma_ions, 'eV', 4) + '\n'
        string += '- Maximum cross section energy:      ' + unit_manager.print_unit(parameters.e_max_sigma_ions, 'eV', 4) + '\n'

    # Print output information
    if print_output:
        if not options.gui_mode: string += '\n'
        string += 'DATA OUTPUT PARAMETERS\n'
        string += 'Verbosity level [0..3]:              ' + str(options.verbosity) + '\n'
        string += 'Python debug level [0..2]:           ' + str(options.debug_lev) + '\n'
        string += 'Fortran debug level [0..2]:          ' + str(options.debug_lev_for) + '\n'
        string += 'Plot cross sections:                 ' 
        if options.plot_xsec: string += 'YES' + '\n'
        else:                 string += 'NO' + '\n'
        string += 'Save data to file:                   '
        if (parameters.save_delay == 0):
            string += 'never' + '\n'
        elif (parameters.save_delay == 1):
            string += 'every cycle' + '\n'
        else:
            string += 'every ' + str(parameters.save_delay) + ' cycles' + '\n'
        if (print_filenames and (parameters.save_delay > 0) ):
            string += '\n'
            string += 'Filename electron mean data:         ' + '\"' + parameters.filename_stat_ele    + EXT +'\"\n'
            string += 'Filename ion mean data:              ' + '\"' + parameters.filename_stat_neu    + EXT + '\"\n'
            string += 'Filename eedf:                       ' + '\"' + parameters.filename_distrib_ele + EXT + '\"\n'
            string += 'Filename iedf:                       ' + '\"' + parameters.filename_distrib_ion + EXT + '\"\n'
            string += 'Filename electric current:           ' + '\"' + parameters.filename_I           + EXT + '\"\n'
            string += 'Filename electric potential:         ' + '\"' + parameters.filename_V           + EXT + '\"\n'
            
        if options.gui_mode:
            string += 'Max number of points in plots:       ' + str(parameters.n_max_points) + '\n'
            string += 'Decimation factor:                   ' + str(parameters.decimation_factor) + '\n'
                        
    return string


def print_gas_information(neutrals):
    string = ''
    for i in range(neutrals.types):
        string += '\n' 
        string += 'Name:                                     ' + str(neutrals.names[i]) + '\n'
        if   (neutrals.molecule_type[i]=='a'):
            string += 'Type:                                     ' + 'atom' + '\n'
        elif (neutrals.molecule_type[i]=='m'):
            string += 'Type:                                     ' + 'non polar molecule' + '\n'
        elif (neutrals.molecule_type[i]=='p'):
            string += 'Type:                                     ' + 'polar molecule' + '\n'
        string +=('Flow rate percentage:                     '
                  + str( unit_manager.fix_digits(neutrals.gas_flow[i]/neutrals.gas_flow.sum()*100.0, 3) )
                  + ' %' + '\n')
        string +=('Partial pressure [Pa]:                    '
                  + str( unit_manager.fix_digits(neutrals.partial_pressure[i], 3) ) + '\n')
        string +=('Number density [m**-3]:                   ' + str( unit_manager.fix_digits(neutrals.number_density[i], 4) )
                                                               + '\n')
        string += 'Mass [u]:                                 ' + str( neutrals.mass[i] ) + '\n'
        string += 'Mean molecule speed [m s**-1]             ' + str( unit_manager.fix_digits(neutrals.mean_v[i], 4) ) + '\n'   
        string +=('Electron energy loss (elastic):           ' + str( unit_manager.fix_digits(neutrals.energy_loss[i], 6) )
                                                               + '\n')
        string += 'Secondary emission coefficient            ' + str( neutrals.secondary_emission[i] ) + '\n'
        string += 'Ionization energy [eV]:                   ' + str( neutrals.ionization_energy[i] ) + '\n'
        string += 'Number of excitation processes:           ' + str( neutrals.excitation_types[i] ) + '\n'
        string += 'Excitation energies [eV]                  '
        for j in range(neutrals.excitation_types[i]):
            string += str(neutrals.excitation_energy[i][j])
        string += '\n'

        if (neutrals.molecule_type[i] != 'a'):
            string += 'Number of dissociation processes:         ' + str( neutrals.dissociation_types[i] ) + '\n'
            string += 'Dissociation energies [eV]                ' 
            for j in range(neutrals.dissociation_types[i]):
                string += str(neutrals.dissociation_energy[i][j])
            string += '\n'
            
        #string += '\nTotal number density relative error = ' + \
        #           str( (neutrals.number_density.sum() - neutrals_density) / neutrals_density ) + '\n'

    return string


def print_runtime_info(charges, neutrals, ccp, parameters, options, time_before, electric_bias_before, n_active_el_before):

    # N. of char allocated to place gas names
    gas_length = 10 
    
    text = ''
#   text += 'Elapsed time before iteration            = ' + unit_manager.print_unit(time_before, 's', 4) + '\n'
    text += 'Iteration time                           = ' + unit_manager.print_unit(parameters.dt_output, 's', 3) + '\n'
    text += 'Timestep                                 = ' + unit_manager.print_unit(charges.dt, 's', 3) + '\n'
    text += 'Computational electrons before cycle     = ' + str(n_active_el_before)
    text += ' (max ' + str((parameters.Nmax_particles))+')' + '\n'
    text += 'Electric bias intensity before cycle     = ' + unit_manager.print_unit(electric_bias_before,'V',3)
    text += ' (max = ' + unit_manager.print_unit(ccp.V_peak, 'V', 3) + ')' + '\n'
    text += 'Reduced electric field before cycle      = ' + unit_manager.print_unit(electric_bias_before
                                                                                   /ccp.distance
                                                                                    /parameters.neutrals_density*1.0e21,'Td',3) 
    text += ' (max = ' + unit_manager.print_unit(ccp.V_peak/ccp.distance/parameters.neutrals_density*1.0e21,'Td',3) + ')\n'
    text += '...' + '\n'
    text += 'Elapsed time after cycle                 = ' + unit_manager.print_unit(charges.time, 's', 4)
    if (parameters.sim_duration > 0):
        text +=  ' of ' + unit_manager.print_unit(parameters.sim_duration, 's', 4)                
        text += ' (' + str('%3.1f' % (100.0*(charges.time)/parameters.sim_duration)) + '%) '
    else:
        text += ' (unlimited)'
    text += '\n'

    text += 'Average time between collisions          = ' + unit_manager.print_unit(charges.tau_mis,'s',3) + '\n'
    text += 'Average current density                  = ' + unit_manager.print_unit(ccp.average_current[0]/ccp.area,'A/m**2',3)\
                                                      + '\n'
    charge_density = 0.0
    for i in range(charges.types):
        if (i==0): charge_density = charge_density + charges.number_density[i] 
        else:      charge_density = charge_density - charges.number_density[i]
    charge_density = charge_density * ELECTRON_CHARGE #ELECTRON_CHARGE is negative
    text += 'Net charge density                       = ' + unit_manager.print_unit(charge_density,'C*m**-3',3) + '\n'
    text += 'Real electron collisions                 = ' + unit_manager.print_exp(neutrals.collisions_total_electron, 3) + \
                                                            str(' (%2.2f' % charges.p_coll) + '%) ' + '\n'
    text += 'Null electron collisions                 = ' + unit_manager.print_exp(neutrals.collisions_null[0], 3) + '\n'
    if True: #(options.verbosity > 2):
        text += '\n'
        text += 'Number of computational particles (max '+str(parameters.Nmax_particles) + \
                ') and physical parameters for electrons and ions' + '\n'
        for i in range(charges.types):
            string = ''
            string += charges.names[i].ljust(gas_length) + ':'
            string += '  N = '        + str(charges.n_active(i)).rjust(8) + ';'
            #string += '  w = '        + unit_manager.print_exp(charges.weight[i],2, align=True).rjust(8) + ';'
            string += '  w = '        + unit_manager.print_exp(charges.weight[i],2, align=True).rjust(5) + ';' 
            string += '  n = '        + unit_manager.print_exp(charges.number_density[i],
                                                               2, align=True).rjust(8) + ' m**-3'  + ';'
            string += '  alpha = '    + unit_manager.print_exp(charges.number_density[i]/parameters.neutrals_density,
                                                               2, align=True).rjust(8) + ';'
            string += '  <E> = '      + unit_manager.print_unit(charges.e_average(i), 'eV',
                                                                3, align=True).ljust(9) + ';'
#            string += '  E/N = '      + unit_manager.print_exp(ccp.V/ccp.distance/parameters.neutrals_density,
#                                                               2, align=True).rjust(8) + ' Td'  + ';'       
            string += '  lD = '       + unit_manager.print_unit(charges.debye_length[i],     'm',
                                                                3, align=True).ljust(9) + ';' 
            string += '  ND = '       + unit_manager.print_exp( 4.0/3.0**math.pi*charges.debye_length[i]**3*\
                                                                            charges.number_density[i], 2,
                                                                            align=True ).rjust(8) + ';'            
            if (charges.debye_length[i] >0):
                    string += '  f = '        + unit_manager.print_unit(charges.v_average(i)/charges.debye_length[i],
                                                                        'Hz', 3, align=True).rjust(9)
            else:
                    string += '  f = '        + 'N/A'.rjust(9)
            if (charges.v_average(i) > 0):
                    string += ' ('            + unit_manager.print_unit(charges.plasma_frequency[i], 'Hz',
                                                                        3, align=True) + ');'
            else:
                    string += '  t = '        + 'N/A'
            text += string + '\n'     

        # Print table of collision rates
            
        text += '\n'
        text += 'Electron collisions' + '\n'
        string = ''
        length = 14
        string += ( 'Gas'.ljust(gas_length)
                    + '| Total'.ljust(length+1)
                    + '| Elastic'.ljust(length+1)
                    + '| Ionization'.ljust(length+1)
                    + '| Excitation'.ljust(length+1)
                    + '| Dissociation'.ljust(length+1)
                    + '| Recombination'.ljust(length+1)                    
                   )

#        for j in range(MAX_DISSOCIATION_TYPES):
#                string += '| Type' + str(j).rjust(2) + '  '
        text += string + '|' + '\n'   
#        string_line = ''.ljust(gas_length+1 + (length+1)*6 + 10*MAX_DISSOCIATION_TYPES, '-')
        string_line = ''.ljust(gas_length+1 + (length+1)*6, '-')
        text += string_line + '\n'
        text += ( 'TOT'.ljust(gas_length) + '|'
                 + unit_manager.print_exp(neutrals.collisions_total_electron,2).rjust(length)      + '|'
                 + unit_manager.print_exp(neutrals.collisions_elastic.sum(),2).rjust(length)       + '|'
                 + unit_manager.print_exp(neutrals.collisions_ionization.sum(),2).rjust(length)    + '|'
                 + unit_manager.print_exp(neutrals.collisions_excitation.sum(),2).rjust(length)    + '|'
                 + unit_manager.print_exp(neutrals.collisions_dissociation.sum(),2).rjust(length)  + '|'
                 + unit_manager.print_exp(neutrals.collisions_recombination.sum(),2).rjust(length) + '|'                  
                 + '\n' )
        text += string_line + '\n'

        for i in range(neutrals.types):
            string = ''
            string += str(neutrals.names[i]).ljust(gas_length) + '|'
            string += unit_manager.print_exp( ( neutrals.collisions_elastic[i]
                                                + neutrals.collisions_ionization[i]
                                                + neutrals.collisions_excitation[i].sum()
                                                + neutrals.collisions_dissociation[i].sum()
                                                + neutrals.collisions_recombination[i].sum()) ,
                                                2 ).rjust(length) + '|'
            string += unit_manager.print_exp( neutrals.collisions_elastic[i],              2 ).rjust(length) + '|'
            string += unit_manager.print_exp( neutrals.collisions_ionization[i],           2 ).rjust(length) + '|'
            string += unit_manager.print_exp( neutrals.collisions_excitation[i].sum(),     2 ).rjust(length) + '|'
            if (neutrals.molecule_type[i] == 'a'):
                string += '-'.rjust(length) + '|'
            else:
                string += unit_manager.print_exp(neutrals.collisions_dissociation[i].sum(), 2).rjust(length)  + '|'
#            for k in range(MAX_DISSOCIATION_TYPES):
#                if ( (neutrals.molecule_type[i] == 'a') or (k >= neutrals.dissociation_types[i]) ):
#                    string += '-'.rjust(9) + '|'
#                else:
#                    string += unit_manager.print_exp(neutrals.collisions_dissociation[i][k],2).rjust(9) + '|'
            string += unit_manager.print_exp( neutrals.collisions_recombination[i].sum(),  2 ).rjust(length) + '|'                    
            text += string + '\n'

        # Print table of dissociation rates and rate constants
        
        text += '\n'
        text += 'Dissociation rate parameters' + '\n'
        string = ''
        length = 20
        string += 'Gas'.ljust(gas_length)
        for j in range(MAX_DISSOCIATION_TYPES):
            string += (  ('| Diss. rate '+str(j)).ljust(length+1)
                       + ('| Diss. rate const. '+str(j)).ljust(length+1) )
        text += string + '|' + '\n'   
        string_line = ''.ljust(gas_length+1 + (length+1)*2*MAX_DISSOCIATION_TYPES , '-')
        text += string_line + '\n'        
        string = 'Unit'.ljust(gas_length)
        for j in range(MAX_DISSOCIATION_TYPES):
            string += (   '| m**-3 * s**-1'.ljust(length+1)
                        + '| m**3  * s**-1'.ljust(length+1)  )
        text += string + '|' + '\n'
        text += string_line + '\n'

        for i in range(neutrals.types):
            string = ''
            # If this gas is atomic, no dissociation is possible, skip it
            if (neutrals.molecule_type[i] != 'a'):
                string += str(neutrals.names[i]).ljust(gas_length) + '|'
                for j in range(MAX_DISSOCIATION_TYPES):
                    if (j >= neutrals.dissociation_types[i]):
                        string += '-'.rjust(length) + '|'
                        string += '-'.rjust(length) + '|'                        
                    else:
                        string += unit_manager.print_exp(neutrals.dissociation_rate[i][j],2).rjust(length) + '|'
                        k = neutrals.dissociation_rate_constant[i][j]
                        if ( isnan(k) or isinf(k) ): 
                            string += 'N/A'
                        else:
                            string += unit_manager.print_exp(k,2).rjust(length) + '|'
                #string += '|'
                text += string + '\n'                    

    return text



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
	Tools for the simulation of plasma discharges: definition of particles ensambles and simulation of particle motion

        The modules in this subpackage define classes to store data about electrons, ions and neutrals.
        They also provide functions which call fortran functions compiled into the fortran package 
        to simulate the motion and interaction of these particles inside plasma discharges,
 
       	* modules
		- reactors 	         -> defines characteristics of plasma reactors, such as electrodes dimensions, bias, etc...
                - moving_particles       -> defines the ensables of electrons and ions
		- target_particles       -> defines neutral particles and their cross sections for impact with electrons and ions

                - particles_data_manager -> read properties of neutral particles from a ascii file
                - particle_mover         -> provides functions to simulate the particles motion and interaction in a cold plasma
                                            it uses the modules contained in the fortran package
"""



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" The _PlotWindow class, used to define a plot window. """

from subprocess import Popen, PIPE, DEVNULL, TimeoutExpired
from os import remove, path, mkdir

from .global_variables import *
from .errors import *
from .funcutils import *


# +---------+
# | Classes |
# +---------+
       
class _PlotWindow:
    """ This class defines a single plot window. """
    
    def __init__(self,
                 gnuplot_default=False,
                 terminal=DEFAULT_TERM,
                 width=DEFAULT_WIDTH,
                 height=DEFAULT_HEIGHT,
                 xpos=DEFAULT_XPOS,
                 ypos=DEFAULT_YPOS,
                 options=None,
                 plot_type=DEFAULT_TYPE,
                 title=None,
                 persistence=PERSISTENCE,
                 redirect_output=REDIRECT_OUT,
                 purge=PURGE_DATA):
        """ An instance of this class contains a plot window.

            Parameters
            ----------
            gnuplot_default: if it is set to True, the terminal 
                             type, window size and position,
                             are not set, leaving the gnuplot defaults 
                             (which are *not* the default values 
                             of the parameters in this method)
            terminal:        terminal type, must be one of
                             the strings in the TERMINALS tuple
            width:           width of the plot window
            height:          height of the plot window
            xpos:            x position of the plot on the screen
            ypos:            y position of the plot on the screen
            options:         a string containing other options for the terminal
            plot_type:       a string defining the type of plot
                             must be one of the strings in PLOT_TYPES
            title:           a string used as the title of the plot
            persistence:     if True, the window on the screen will not close 
                             after the gnuplot process is terminated
            redirect_output: True:  save gnuplot otuput and errors to files
                             False: send then to /dev/stdout and /dev/stderr
                             None:  send them to /dev/null
            purge:           if True, old datafiles are deleted each time 
                             new data is plotted on this window

            Initialized data attributes
            ---------------------------

            self.window_number:   integer number that identifies the plot window
            self.gnuplot_process: gnuplot process (instance of subprocess.Popen)
            self.term_type:       the type of gnuplot terminal started 
                                  (string, e.g. 'x11' or 'wxt')
            self.persistence:     True if the window is persistent
            self.filename_out:    file to which gnuplot output is redirected 
            self.filename_err:    file to which gnuplot error messages
                                  are redirected
            self.plot_type:       a string defining the type of plot
            self.n_axes:          number of plot axes (2 for 2D plots, 3 for 3D ones)
            self.xmin:            minimum of x-axis range (initialized to None)
            self.xmax:            maximum of x-axis range (initialized to None)
            self.ymin:            minimum of y-axis range (initialized to None)
            self.ymax:            maximum of y-axis range (initialized to None)
            self.zmin:            minimum of z-axis range (initialized to None)
            self.zmax:            maximum of z-axis range (initialized to None)
            self.title:           the window title (None if not given)
            self.data_filenames:  list containing the names of the data files
            self.functions:       list containing the functions plotted
            self.n_volatiles      number of curves plotted as volatile
                                  (passing the data inline, without writing to file)
            self.purge:           if True, old datafiles are deleted each time 
                                  new data is plotted
            self.error:           one of the tuples in the errors.py module
        """

        self.error = NOERROR
        
        # Check that terminal and plot type are among the known ones;
        # otherwise, use defaults and store an error message in self.error
        if terminal in TERMINALS:
            self.term_type = terminal
        else:
            self.term_type = DEFAULT_TERM
            (status, message) = ERROR_UNKNOWN_TERM
            message += (' \"' + str(terminal) + '\"'
                        + ', using default  \"'
                        + DEFAULT_TERM + '\"')
            self.error = (status, message)
        if plot_type in PLOT_TYPES: 
            self.plot_type = plot_type
        else:
            self.plot_type = DEFAULT_TYPE
            (status, message) = ERROR_UNKNOWN_TYPE
            message += (' \"' + str(plot_type) + '\"'
                        + ', using default \"'
                        + DEFAULT_TYPE + '\"' )  
            self.error = (status, message)

        # Store the persistence status
        self.persistence = persistence

        # Store the purge status
        self.purge = purge
        
        # Store the window title (may be None)
        self.title = title        

        # Store the number of axis     
        if (self.plot_type == '3D'): self.n_axes = 3
        else:                        self.n_axes = 2
        
        # Assign a unique number to this plot window
        # to be used to create unique names for datafiles
        self.window_number = None
        for i in range(len(window_list)):
            if (i != window_list[i].window_number):
                self.window_number = i
                break
        if (self.window_number is None): self.window_number = len(window_list)
        
        # Add this plot window to the list of the active ones
        window_list.append(self)
        
        # Initialize the lists where data filenames 
        # and functon strings will be stored
        self.data_filenames = []
        self.functions      = []
        self.n_volatiles    = 0

        # Initialize the axis ranges
        self.xmin = None
        self.xmax = None
        self.ymin = None
        self.ymax = None
        self.zmin = None
        self.zmax = None            
                
        # Start the gnuplot process to plot on this window
        if persistence: exec_list = ['gnuplot','-p']
        else:           exec_list = ['gnuplot']        
        if redirect_output is True:
            # If redirection of gnuplot output to files is requested
            # Create the directories to save gnuplot output and errors
            if not path.exists(DIRNAME):     mkdir(DIRNAME)
            if not path.exists(DIRNAME_OUT): mkdir(DIRNAME_OUT)
            # Create the filenames
            if not path.exists(DIRNAME_ERR): mkdir(DIRNAME_ERR)
            self.filename_out = path.join( DIRNAME_OUT, FILENAME_OUT
                                           + '_w_' + str(self.window_number) )
            self.filename_err = path.join( DIRNAME_ERR, FILENAME_ERR
                                           + '_w_' + str(self.window_number) )
            if self.title is not None:
                self.filename_out += '(' + correct_filename(self.title) + ')'
                self.filename_err += '(' + correct_filename(self.title) + ')'
            # Start the gnuplot process with output redirected to files
            self.gnuplot_process = Popen(exec_list,
                                         stdin=PIPE,
                                         stdout=open(self.filename_out, 'w'),
                                         stderr=open(self.filename_err, 'w'),
                                         universal_newlines=True)
        elif redirect_output is False:
            # Start the gnuplot process without redirection
            self.filename_out = '/dev/stdout'
            self.filename_err = '/dev/stderr'               
            self.gnuplot_process = Popen(exec_list,
                                         stdin=PIPE,
                                         universal_newlines=True)            
        else:
            # Start the gnuplot process silenced
            self.filename_out = '/dev/null'
            self.filename_err = '/dev/null'        
            self.gnuplot_process = Popen(exec_list,
                                         stdin=PIPE,
                                         stdout=DEVNULL,
                                         stderr=DEVNULL,
                                         universal_newlines=True)
        if not gnuplot_default:
            command_string = ( 'set terminal ' + str(self.term_type)
                               + ' size '      + str(width) + ',' + str(height)
                               + ' position '  + str(xpos)  + ',' + str(ypos)   )
            if (options is not None):
                command_string += ' ' + str(options)
            self._command(command_string)
        if title is not None:
            self._command('set title \"' + correct_filename(self.title) + '\"')
            
            
    def _command(self, command_string):
        """ Send a command to the gnuplot process.           

            Parameters
            ----------

            command_string: string containing a gnuplot command
                            NOTE: no check is made that the string 
                            is a valid gnuplot command
        """

        self.gnuplot_process.stdin.write(command_string + EOL)
        self.gnuplot_process.stdin.flush()

        
    def _quit_gnuplot(self):
        """ Terminate the gnuplot process associated to this window. 

            Returns
            -------

            The last output provided by the gnuplot process
        """

        try:
            gnuplot_last_output = self.gnuplot_process.communicate(
                input='quit', timeout=TIMEOUT_QUIT)
        except TimeoutExpired:
            self.gnuplot_process.kill()
            gnuplot_last_output = self.gnuplot_process.communicate()

        return gnuplot_last_output                    
        

    def _add_functions(self, function_list, replot=False):
        """ Add one or more functions to a plot window.

            Parameters
            ----------

            function_list: a list in the following form:
                          [ [function1, label1, options1], 
                            [function2, label2, options1], ... ]
                          where:
                          - function1 is a string defining the function 
                            to be plotted
                            e.g. '2*x**2+3*cos(x)' or 'sin(x**2+y**2) 
                          - label1 is a string, that will be used to identify 
                            the plot in the legend, or None
                          - options1 is a string, containing additional options
                            or None 
            replot:       if True, the functions are plotted without erasing 
                          previously plotted ones
            Returns
            -------
            
            One of the tuples defined in the errors.py module
        """

        (status, message) = NOERROR

        if (self.n_volatiles and replot): return ERROR_REPLOT_VOLATILE
        if not function_list: return ERROR_NO_FUNCTION 
        if (replot and (not self.data_filenames) and (not self.functions)):
            return ERROR_NO_REPLOT
        # Check consistency of the whole list
        for i in range(len(function_list)):
            if ( len(function_list[i]) != 3):
                (status, message) = ERROR_PLOT_PARAMETERS
                if len(function_list) > 1:
                    message += ' in list item # ' + str(i)
                message += ': 3 items expected, given ' + str(len(function_list[i]))
                return (status, message)
                   
        # If it is not a replot, reset the plot window   
        if not replot:
            # If purge status is True, remove old datafiles
            if self.purge:
                for filename in self.data_filenames:
                    try:
                        remove(filename)
                    except FileNotFoundError:
                        pass
            # Clear the list of datafiles and functions
            # and reset the number of volatiles
            self.data_filenames.clear()
            self.functions.clear()
            self.n_volatiles = 0
            
        # Beginning of the command string to send to gnuplot
        if replot:                     command_string = 'replot '
        elif (self.plot_type == '3D'): command_string = 'splot '
        else:                          command_string = 'plot '

        command_list = []
              
        for function_item in function_list:
            # Initialize the command string for this fucntion
            function_command_string = ''
            # Add the function to the function list
            self.functions.append(str(function_item[0]))
            
            # Read the label, if given
            if (function_item[1] is None): label = None
            else:                          label = str(function_item[1])
            
            # Add  to the command string the plot command for this function
            function_command_string += str(function_item[0])
            if (label is not None):
                function_command_string += ' title \"' + str(label) + '\"'
            if (function_item[2] is not None):
                function_command_string += ' ' + str(function_item[2])
                
            command_list.append(function_command_string)

        command_string += ', '.join(command_list)
                        
        # Send command string to gnuplot   
        self._command(command_string)

        return status, message        
        

    def _add_curves(self, data_list, volatile=VOLATILE, replot=False):
        """ Add one or more curves from data to the plot window. 

            Parameters
            ----------

            data_list:    a list in one of the following forms:
                          2D data: [ [x1 ,  y1, label1, options1],     
                                     [x2,   y2, label2, options2],... ]
                          3D data: [ [x1,   y1, z1, label1, options1], 
                                     [x2,   y2, z2, label2, options2], ... ] 
                          where:
                          - x1 contains the x-coordinates of the points to plot;
                            for 2D plot windows this can be set to None, 
                            which is useful if the data are 1D, as in the case 
                            of boxplots, or if you want gnuplot to 
                            automatically provide x-values.
                            2D plots with and without x values can be mixed:
                            [ [x1,   y1, label1, options1],     
                              [None, y2, label2, options2],... ]  
                          - y1 contains the y-coordinates of the points to plot 
                          - z1 contains the z-coordinates of the points to plot  
                            (for 3D plot windows only)
                          - label1 is a string, that will be used to identify the plot 
                            in the legend, or None
                          - options1 is a string, containing additional options
                            or None
                          The form must be consisted with the type of plot 
                          that was defined at the plot window creation, 
                          otherwise an error message is returned
            volatile:     if True, the data are not written to file, but sent 
                          to gnuplot as volatile data using the special filename '-'
            replot:       if True, the curves are plotted without erasing 
                          previously plotted ones

            Returns
            -------

            One of the tuples defined in the errors.py module
        """

        (status, message) = NOERROR

        if (self.n_volatiles and replot): return ERROR_REPLOT_VOLATILE      
        if not data_list: return ERROR_NO_DATA
        if (replot and (not self.data_filenames) and (not self.functions)):
            return ERROR_NO_REPLOT
        
        # Check consistency of the each data list item
        for i in range(len(data_list)):
            # Check the number of data given
            if (len(data_list[i]) != self.n_axes + 2):
                (status, message) = ERROR_PLOT_PARAMETERS
                if len(data_list) > 1: message += ' in list item # ' + str(i)
                message += ( ': ' +  str(self.n_axes + 2)
                             + ' items expected, given '
                             + str(len(data_list[i])) )
                return (status, message)            
            
            # Check if xdata are missing in this item            
            if (data_list[i][0] is None):
                xmissing = True
                len_x = None
                # Missing x-data are not allowed in 3D plots
                if (self.plot_type == '3D'):
                    (status, message) = ERROR_WRONG_TYPE
                    if (len(data_list) > 1): message += ' in list item # ' + str(i)
                    message += ': missing x-data in a 3D plot window'
                    return status, message
            else:
                xmissing = False
                
            # Transform single numbers to lists
            if not xmissing:
                try:
                    len_x = len(data_list[i][0])
                except TypeError:
                    data_list[i][0] = [ data_list[i][0] ]
                    len_x = 1
            try:
                len_y = len(data_list[i][1])
            except TypeError:
                data_list[i][1] = [ data_list[i][1] ]
                len_y = 1
            if (self.n_axes == 3):
                # Transform single z-numbers to lists
                try:
                    len_z = len(data_list[i][2])
                except TypeError:
                    data_list[i][2] = [ data_list[i][2] ]
                    len_z = 1
                    
            # Check consistency of the data lengths
            if ( (not xmissing) and (len_x != len_y) ):
                (status, message) = ERROR_PLOT_PARAMETERS
                if len(data_list) > 1:  message += ' in list item # ' + str(i)
                message += ': x and y data have different sizes'
                return status, message
            elif ( (self.n_axes == 3) and (len_z != len_y) ):
                (status, message) = ERROR_PLOT_PARAMETERS
                if len(data_list) > 1:  message += ' in list item # ' + str(i)
                message += ': y and z data have different sizes'
                return status, message               
            elif (len_y == 0):
                (status, message) = ERROR_PLOT_PARAMETERS
                if len(data_list) > 1: message += ' in list item # ' + str(i)
                message += ': zero size data given'
                return status, message

        # Create the data files directory, if it doesn't exist,
        # unless a volatile plot has been requested
        if not volatile:
            if not path.exists(DIRNAME):      mkdir(DIRNAME)
            if not path.exists(DIRNAME_DATA): mkdir(DIRNAME_DATA)
                   
        # If it is not a replot, reset the plot window   
        if not replot:
            # If purge status is True, remove old datafiles
            if self.purge:
                for filename in self.data_filenames:
                    try:
                        remove(filename)
                    except FileNotFoundError:
                        pass
            # Clear the list of datafiles and functions
            # and reset the number of volatiles
            self.data_filenames.clear()
            self.functions.clear()
            self.n_volatiles = 0            
        
        # Beginning of the command string to send to gnuplot
        if replot:                     command_string = 'replot '
        elif (self.plot_type == '3D'): command_string = 'splot '
        else:                          command_string = 'plot '

        # Initialize the list in which commands for each curve
        # will be stored
        command_list = []
        
        # Save the old number of curves before adding the new ones
        n_curves = len(self.data_filenames)       
        for data_item in data_list:
            # Initialize the string for this curve
            curve_command_string = ''
            
            # Read the curve label, if given
            if (data_item[self.n_axes] is None):
                label = None
            else:
                label = str(data_item[self.n_axes])

            # Read additional options, if given
            if (data_item[self.n_axes+1] is None):
                options = None
            else:
                options = str(data_item[self.n_axes+1])
                        
            # Define the unique filename for the data file
            # or use the gnuplot special filename '-' for volatile data
            if volatile:
                filename = FILENAME_VOLATILE
            else:
                curve_number = n_curves + i
                name_string = FILENAME_DATA 
                name_string += '_w' + str(self.window_number)
                if (self.plot_type == '3D'): name_string += '_3D'
                elif (data_item[0] is None): name_string += '_1D'
                else:                        name_string += '_2D'        
                if self.title is not None:
                    name_string += '(' + correct_filename(self.title) + ')'
                name_string += '_c' + str(curve_number)
                if label is not None:
                    name_string += '(' + correct_filename(label) + ')' 
                name_string += FILENAME_DATA_EXT
                filename = path.join( DIRNAME_DATA, name_string )

            # Add filename to list, or increase volatile data counter
            if volatile:
                self.n_volatiles += 1
            else:
                self.data_filenames.append(filename)

            # Save data to data files, unless they are volatile
            if not volatile:
                if (self.plot_type == '3D'):
                    data_file_3d( data_item[0],
                                  data_item[1],
                                  data_item[2],
                                  filename )

                elif (data_list[i][0] is None):
                    data_file_1d( data_item[1],
                                  filename )

                else:
                    data_file_2d( data_item[0],
                                  data_item[1],
                                  filename )

            # Add plot command to the command list
            curve_command_string += '\"' + filename + '\"'
            if (label is None):
                # If we don't set a title, gnuplot automatically
                # uses the data file name, so we set "" as title
                curve_command_string += ' title \"\"'
            else:
                curve_command_string += ' title \"' + label + '\"'
            if (options is not None):
                curve_command_string += ' ' + options
                
            command_list.append(curve_command_string)

        command_string += ', '.join(command_list)

        # For volatile data, add the datastrings          
        if volatile:
            command_string += EOL
            for data in data_list:
                if (self.plot_type == '3D'):
                    command_string += data2string_3d(data[0],
                                                   data[1],
                                                   data[2])

                elif (data[0] is None):
                    command_string += data2string_1d(data[1])

                else:
                    command_string += data2string_2d(data[0],
                                                   data[1])
                        
        # Send command string to gnuplot   
        self._command(command_string)

        return status, message





# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Physical constants used in this package.

    This module contains physical constants required for use in the package
        
    Documentation is also available in the docstrings.
"""

# +--------------------+
# | Physical constants |
# +--------------------+

K_BOLTZMANN             = 1.3806503E-23         # / J*K**-1
EPSILONZERO             = 8.854187818E-12       # / F*m**-1
ELECTRON_CHARGE         = -1.602176E-19         # / C
ELECTRON_MASS           = 9.109382E-31          # / kg
ATOMIC_UNIT_MASS        = 1.66054E-27           # / kg
ZERO_CELSIUS            = 273.15                # / K           
ATM_PRESSURE_PA         = 1.0133E5              # / Pa          atmospheric pressure in pascal
ATM_PRESSURE_TORR       = 760.0                 # / torr        atmospheric pressure in Torr
ATM_PRESSURE_PSI        = 14.696                # / psi         atmospheric pressure in psi



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Some utility functions. """

from .global_variables import *


def correct_filename(filename):
    """ Remove invalid characters from a filename.

        If the string contains characters in INVALID_CHARS, they are
        substituted with the char in SUBSTITUTE_CHAR.

        Parameters
        ----------

        filename: a string, representinge a filename

        Returns
        -------

        The modified string
    """

    for c in INVALID_CHARS:
        filename = filename.replace(c, SUBSTITUTE_CHAR)

    return filename


def data2string_1d(data, eol=EOL):
    """ Create a string with the given data, to be sent to gnuplot.

        The string can be used to pass data to gnuplot using the
        special filename "-". The form of the string is
        "data[0]\ndata[1]\n ... data[len(data)]\ne\n"

        Parameters
        ----------

        data:  data representing the points to plot
        eol:   end-of-line character

        Returns
        -------
 
        A string with the data in the form requested by gnuplot
    """
    #s = ''
    #for value in data:
    #    s += str(value) + eol
    #s += 'e' + eol
    #return s
    
    l = []
    for value in data:
        l.append(str(value))
    l.append('e')
    return eol.join(l) + eol


def data2string_2d(x_data, y_data, eol=EOL):
    """ Create a string with the given data, to be sent to gnuplot.

        The string can be used to pass data to gnuplot using the
        special filename "-". The form of the string is
        "x_data[0] y_data[0]\nx_data[1] y_data[1]\n ... 
         ... x_data[len(data)] y_data[len(data)]\ne\n"

        Parameters
        ----------

        x_data:   data representing x coordinates of the points to plot
        y_data:   data representing y coordinates of the points to plot
        eol:      end-of-line character

        Returns
        -------
 
        A string with the data in the form requested by gnuplot
    """
    #s = ''
    #for i in range(len(x_data)):
    #    s += str(x_data[i]) + ' ' + str(y_data[i]) + eol       
    #s += 'e' + eol    
    #return s

    l = []
    for i in range(len(x_data)):
        l.append( ' '.join( (str(x_data[i]), str(y_data[i])) ) )
    l.append('e')
    return eol.join(l) + eol
               

def data2string_3d(x_data, y_data, z_data, eol=EOL):
    """ Create a string with the given data, to be sent to gnuplot.

        The string can be used to pass data to gnuplot using the
        special filename "-". The form of the string is
        "x_data[0] y_data[0] z_data[0]\nx_data[1] y_data[1]  z_data[1]\n ... 
         ... x_data[len(data)] y_data[len(data)] z_data[len(data)]\ne\n"

        Parameters
        ----------

        x_data:   data representing x coordinates of the points to plot
        y_data:   data representing y coordinates of the points to plot
         z_data:   data representing z coordinates of the points to plot
        eol:      end-of-line character

        Returns
        -------
 
        A string with the data in the form requested by gnuplot
    """
    #s = ''
    #for i in range(len(x_data)):
    #    s += str(x_data[i]) + ' ' + str(y_data[i]) + ' ' + str(z_data[i]) + eol       
    #s += 'e' + eol
    #return s
    l = []
    for i in range(len(x_data)):
        l.append( ' '.join( (str(x_data[i]), str(y_data[i]), str(z_data[i])) ) )
    l.append('e')
    return eol.join(l) + eol
    

def data_file_1d(x_data, filename, eol=EOL):
    """ Create a file and write 1d data to it in csv format.

        The file will contain 1d data in ascii format, in a single column:

        x1
        x2
        ..

        Parameters
        ----------

        x_data:   data representing the points to plot
        filename: the name of the data file to be created
        eol:      end-of-line character
    """

    data_file = open(filename, 'w') 
    for x in x_data:
        data_file.write(str(x) + eol)
    data_file.close()


def data_file_2d(x_data, y_data, filename, sep=SEP, eol=EOL):
    """ Create a file and write 2d data to it in csv format.

        The file will contain data in ascii format, in the form:

        x1  y1
        x2  y2
        ..  ..

        Parameters
        ----------

        x_data:   data representing x coordinates of the points to plot
        y_data:   data representing y coordinates of the points to plot
        filename: the name of the data file to be created
        sep:      separator character between colums (vaues in a line)
        eol:      end-of-line character
    """

    data_file = open(filename, 'w') 
    for i in range(len(x_data)):
        data_file.write( str(x_data[i])
                         + sep
                         + str(y_data[i])
                         + eol )
    data_file.close()


def data_file_3d(x_data, y_data, z_data, filename, sep=SEP, eol=EOL):
    """ Create a file and write 3d data to it in csv format.

        The file will contain data in ascii format, in the form:

        x1  y1  z1
        x2  y2  z2
        ..  ..  ..

        Parameters
        ----------

        x_data:   data representing x coordinates of the points to plot
        y_data:   data representing y coordinates of the points to plot
        z_data:   data representing z coordinates of the points to plot
        filename: the name of the data file to be created
        sep:      separator character between colums (vaues in a line)
        eol:      end-of-line character
    """

    data_file = open(filename, 'w') 
    for i in range(len(x_data)):
        data_file.write(  str(x_data[i])
                          + sep
                          + str(y_data[i])
                          + sep
                          + str(z_data[i])                                
                          + eol )
    data_file.close()
